import tensorflow as tf
from typing import Optional
from ospark.nn.loss_function.loss_function import LossFunction


class CrossEntropy(LossFunction):

    def __init__(self, reduction: Optional[str]=None):
        """
        Args
            reduction: Optional[str]
                There are two options are mean and sum, the default is mean.
        """

        self._reduction = reduction or "mean"
        if self._reduction != "mean" and self._reduction != "sum":
            raise AttributeError(f"reduction must be mean or sum, doesn't have {reduction}")

    def calculate(self,
                  prediction: tf.Tensor,
                  target_data: tf.Tensor,
                  mask_matrix: Optional[tf.Tensor]=None) -> tf.Tensor:
        loss_value = target_data * tf.math.log(prediction)
        if mask_matrix is not None:
            loss_value *= mask_matrix
        if self._reduction == "sum":
            return -tf.reduce_sum(loss_value)
        else:
            if mask_matrix is not None:
                elements_number = tf.reduce_sum(mask_matrix)
                return -tf.reduce_sum(loss_value) / elements_number
            else:
                return -tf.reduce_mean(loss_value)


from __future__ import annotations
from ospark.nn.block import Block
from ospark.nn.component.basic_module import ModelObject
from typing import List, NoReturn, Optional
from abc import abstractmethod
import tensorflow as tf

class Cell(ModelObject):

    def __init__(self, obj_name: str, is_training: Optional[bool]=None, training_phase: Optional[bool]=None):
        super().__init__(obj_name=obj_name, is_training=is_training, training_phase=training_phase)

    def pipeline(self, input_data: tf.Tensor) -> tf.Tensor:
        output = input_data
        for block in self.assigned:
            output = block.pipeline(input_data=output)
        return output

    @abstractmethod
    def in_creating(self) -> NoReturn:
        pass

    @classmethod
    def draw_structure(cls, block_classes: List[Block]):
        def set_configs(obj_name: str, configs) -> Cell:
            cell = cls(obj_name=obj_name)
            for block, config in zip(block_classes, configs):
                cell.assign(block(**config))
            return cell
        return set_configs


from __future__ import annotations
from typing import Tuple, Optional, Union
import tensorflow as tf
import math


class DataGenerator:

    class Dataset:

        def __init__(self):
            self._training_data = None
            self._target_data   = None

        @property
        def training_data(self) -> tf.Tensor:
            return self._training_data

        @property
        def target_data(self) -> tf.Tensor:
            return self._target_data

        def data_setting(self, training_data: tf.Tensor, target_data: tf.Tensor):
            self._training_data = training_data
            self._target_data   = target_data


    def __init__(self,
                 training_data: Union[tf.Tensor, list],
                 target_data: Union[tf.Tensor, list],
                 batch_size: int,
                 initial_step: Optional[int]=None):
        self._training_data  = training_data
        self._target_data    = target_data
        self._batch_size     = batch_size
        self._max_step       = math.ceil(len(training_data) / batch_size)
        self._step           = initial_step or 0
        self._dataset        = self.Dataset()

    @property
    def training_data(self) -> tf.Tensor:
        return self._training_data

    @property
    def target_data(self) -> tf.Tensor:
        return self._target_data

    @property
    def batch_size(self) -> int:
        return self._batch_size

    @property
    def step(self) -> int:
        return self._step

    @property
    def max_step(self) -> int:
        return self._max_step

    @property
    def dataset(self) -> Dataset:
        return self._dataset

    def __iter__(self):
        return self

    def __next__(self) -> Dataset:
        if self.step < self.max_step:
            dataset = self._get_data()
            self._step += 1
            return dataset
        self.reset()
        raise StopIteration()

    def reset(self):
        self._step = 0

    def _get_data(self) -> Dataset:
        start = self.step * self.batch_size
        end   = start + self.batch_size

        training_data = self.training_data[start: end]
        target_data   = self.target_data[start: end]

        self.dataset.data_setting(training_data=training_data, target_data=target_data)
        return self.dataset

            


from pathlib import Path
import numpy as np
from typing import Tuple


class DataFolder:

    def __init__(self,
                 train_data_folder: str,
                 label_data_folder: str):
        self._train_data_folder = Path(train_data_folder)
        self._label_data_folder = Path(label_data_folder)
        self._training_files    = list(self._train_data_folder.iterdir())
        self._labeling_files    = list(self._label_data_folder.iterdir())

        self._indexed_training_data = self.data_indexing(self.training_files)
        self._indexed_labeling_data = self.data_indexing(self.labeling_files)

    @property
    def train_data_folder(self) -> Path:
        return self._train_data_folder

    @property
    def label_data_folder(self) -> Path:
        return self._label_data_folder

    @property
    def training_files(self) -> list:
        return [str(file_name).split("/")[-1] for file_name in self._training_files]

    @property
    def labeling_files(self) -> list:
        return [str(file_name).split("/")[-1] for file_name in self._labeling_files]

    @property
    def indexed_training_data(self) -> dict:
        return self._indexed_training_data

    @property
    def indexed_labeling_data(self) -> dict:
        return self._indexed_labeling_data

    def data_indexing(self, files: list) -> dict:
        files.sort()

        filtered_file = [string for string in files if any([char.isdigit() for char in string])]
        datasets = map(self._indexed, filtered_file)
        return dict(datasets)

    def _indexed(self, file_name: str) -> Tuple[int, str]:
        number_index = np.where(np.array([char.isdigit() for char in file_name]) == True)[0]
        index        = int("".join([file_name[i] for i in number_index]))
        return index, file_name

    def get_files(self, index: int) -> Tuple[Path, Path]:
        training_data = self.train_data_folder / self.indexed_training_data[index]
        labeling_data = self.label_data_folder / self.indexed_labeling_data[index]
        return training_data, labeling_data


from __future__ import annotations
from ospark.nn.cell import Cell
from ospark.nn.block.resnet_block import ResnetBlock, BottleneckBuildingBlock, BuildingBlock
from typing import List, NoReturn, Optional
import tensorflow as tf


class ResnetCell(Cell):

    def __init__(self,
                 obj_name: str,
                 blocks: List[ResnetBlock],
                 is_training: Optional[bool]=None,
                 training_phase: Optional[bool]=None):
        super().__init__(obj_name=obj_name, is_training=is_training, training_phase=training_phase)
        self._blocks    = blocks

    @property
    def blocks(self) -> List[ResnetBlock]:
        return self._blocks

    @classmethod
    def build_cell(cls,
                   obj_name: str,
                   block_type: str,
                   block_number: int,
                   scale_rate: int,
                   input_channel: int,
                   main_channel: int,
                   strides: int,
                   is_training: bool
                   ) -> ResnetCell:
        blocks = []
        if block_type == "block_1":
            block_cls = BottleneckBuildingBlock
        elif block_type == "block_2":
            block_cls = BuildingBlock
        else:
            raise KeyError("block_type is only block_1 and block_2, please check block_type")

        for i in range(block_number):
            name = f"block_{i}"
            if i == 0:
                block = block_cls(obj_name=name,
                                  input_channel=input_channel,
                                  main_channel=main_channel,
                                  scale_rate=scale_rate,
                                  strides=strides,
                                  use_shortcut_conv=True,
                                  is_training=is_training)
            else:
                block = block_cls(obj_name=name,
                                  input_channel=input_channel,
                                  main_channel=main_channel,
                                  scale_rate=scale_rate,
                                  strides=1,
                                  is_training=is_training)
            blocks.append(block)
        return cls(obj_name=obj_name, blocks=blocks, is_training=is_training)

    def in_creating(self) -> NoReturn:
        for block in self.blocks:
            self.assign(component=block)

    def pipeline(self, input_data: tf.Tensor) -> tf.Tensor:
        output = input_data
        for block in self.blocks:
            output = block.pipeline(output)
        return output


def resnet_cells(cells_number: List[int],
                 input_channels: List[int],
                 main_channels: List[int],
                 scale_rate: int,
                 block_type: str,
                 is_training: bool) -> List[ResnetCell]:
    cells = []
    for i, parameters in enumerate(zip(cells_number, input_channels, main_channels)):
        block_number, input_channel, main_channel = parameters
        name = f"cell_{i}"
        if i == 0:
            cell = ResnetCell.build_cell(obj_name=name,
                                         block_type=block_type,
                                         block_number=block_number,
                                         scale_rate=scale_rate,
                                         input_channel=input_channel,
                                         main_channel=main_channel,
                                         strides=1,
                                         is_training=is_training
                                         )
        else:
            cell = ResnetCell.build_cell(obj_name=name,
                                         block_type=block_type,
                                         block_number=block_number,
                                         scale_rate=scale_rate,
                                         input_channel=input_channel,
                                         main_channel=main_channel,
                                         strides=2,
                                         is_training=is_training
                                         )
        cells.append(cell)
    return cells




from __future__ import annotations
import tensorflow as tf 
from typing import List, NoReturn, Optional
from .name_space import NameSpace


class Weight:

    def __init__(self, 
                 obj_name: str,
                 shape: List[int],
                 trainable: Optional[bool]=None):
        self._obj_name      = obj_name
        self._indexed_name  = None
        self._value         = None
        self._shape         = shape
        self._scale         = tf.constant(1.0)
        self._name_space    = NameSpace("")
        if trainable is None:
            self._trainable          = True
            self._is_default_setting = True
        else:
            self._trainable          = trainable
            self._is_default_setting = False

    @property
    def obj_name(self) -> str:
        return self._obj_name
    
    @property
    def indexed_name(self) -> None:
        return self._indexed_name

    @property
    def value(self) -> None:
        return self._value

    @property
    def shape(self) -> List[int]:
        return self._shape

    @property
    def trainable(self) -> bool:
        return self._trainable

    def __mul__(self, other):
        if self._value is not None:
            self._value *= other
        else:
            self._scale = tf.constant(other)
        return self

    def __str__(self) -> str:
        if self._value is not None:
            return str(self._value.numpy().tolist())
        else:
            return str("value is None.")

    def init_weight(self) -> tf.Tensor:
        raise NotImplementedError("instance class Weight must implement init_weight.")

    def create(self) -> NoReturn:
        manager = WeightOperator()
        self._indexed_name = self._name_space._prefix.name + self.obj_name
        manager.add_weight(self)
        if self._value is None:
            #TODO debug mode.
            #print(f"Initialize weight {self.indexed_name}.")
            self._value = tf.Variable(self.init_weight() * self._scale, trainable=self.trainable)

    def restore(self, weight: tf.Tensor) -> NoReturn:
        if weight is not None:
            if weight.shape == self.shape:
                self._value = tf.Variable(weight, trainable=self.trainable)
            else:
                print(f"Weight \"{self.indexed_name}\", shape does not match the original setting, so use the initialization weight")


class WeightOperator:
    _instance        = None
    _first_initial   = False
    _restore_weights = {}

    def __new__(cls) -> WeightOperator:
        if not cls._instance:
            WeightOperator._instance = super().__new__(cls)
        return cls._instance

    def __init__(self) -> NoReturn:
        if not self._first_initial:
            self._weights = []
            WeightOperator._first_initial = True

    def add_weight(self, weight: Weight) -> NoReturn:
        self._weights.append(weight)
        if weight.indexed_name in self._restore_weights:
            weight.restore(tf.convert_to_tensor(self._restore_weights.pop(weight.indexed_name), dtype=tf.float32))

    @classmethod
    def restore(cls, weights: dict) -> NoReturn:
        cls._restore_weights.update(weights)

    @classmethod
    def clean_weights(cls) -> NoReturn:
        cls._restore_weights.clear()

    def collect_weights(self, partition_name: Optional[str]=None) -> List[tf.Tensor]:
        if partition_name is None:
            return [weight.value for weight in self._weights if weight.trainable]
        else:
            partition_weight = [weight.value for weight in self._weights if partition_name in weight.indexed_name and weight.trainable]
            if partition_weight == []:
                raise NameError(f"partition name {partition_name} is not exist, please check")
            else:
                return partition_weight

    def collect_weights_name(self, partition_name: Optional[str]=None) -> List[str]:
        if partition_name is None:
            return [weight.indexed_name for weight in self._weights if weight.trainable]
        else:
            partition_weights_name = [weight.indexed_name for weight in self._weights if partition_name in weight.indexed_name and weight.trainable]
            if partition_weights_name == []:
                raise NameError(f"partition name {partition_name} is not exist, please check")
            else:
                return partition_weights_name

    def get_weights(self, partition_name: Optional[str]) -> dict:
        if partition_name is None:
            return self.weights
        else:
            partition_weights_name = [weight.indexed_name for weight in self._weights if partition_name in weight.indexed_name]
            if partition_weights_name == []:
                raise NameError(f"partition name {partition_name} is not exist, please check")
            else:
                return partition_weights_name


    @property
    def weights(self) -> dict:
        return {weight.indexed_name: weight.value.numpy().tolist() for weight in self._weights}

    @classmethod
    def release_loaded_weight(cls) -> NoReturn:
        del cls._restore_weights





import tensorflow as tf
from typing import NoReturn


class LossFunction:

    def __init_subclass__(cls) -> NoReturn:
        super().__init_subclass__()
        setattr(LossFunction, cls.__name__, cls)

    def calculate(self, prediction: tf.Tensor, target_data: tf.Tensor) -> tf.Tensor:
        raise NotImplementedError()

    def __call__(self, prediction: tf.Tensor, target_data: tf.Tensor) -> tf.Tensor:
        return self.calculate(prediction, target_data)





from __future__ import annotations
from ospark.nn.block.connection_block import shared_convolution_decoder
from typing import NoReturn, Union, Optional
from ospark.nn.block import Block
from .backbone import Backbone
from .builder import resnet50
import tensorflow as tf
from typing import List


class Unet(Backbone):

    def __init__(self,
                 obj_name: str,
                 downsampling: Union[Block, Backbone],
                 upsampling: Block):
        super().__init__(obj_name=obj_name,
                         is_keep_blocks=False,
                         is_training=True)
        self._downsampling = downsampling
        self._upsampling   = upsampling

    @property
    def downsampling(self) -> Backbone:
        return self._downsampling

    @property
    def upsampling(self) -> Block:
        return self._upsampling

    def in_creating(self) -> NoReturn:
        self.assign(component=self.downsampling, name="downsampling")
        self.assign(component=self.upsampling, name="upsampling")

    def pipeline(self, input_data: tf.Tensor) -> tf.Tensor:
        encoder_output = self.downsampling.pipeline(input_data)
        if type(encoder_output) is not list:
            raise TypeError("Type is not list, please check use_catch of backbone attribute")
        input_data       = encoder_output.pop()
        connection_input = encoder_output
        decoder_output = self.upsampling.pipeline(input_data=input_data, connection_input=connection_input)
        return decoder_output

    @classmethod
    def build_shared_conv(cls,
                          input_channels: Optional[List[int]]=None,
                          output_channels: Optional[List[int]]=None,
                          is_training: Optional[bool]=None) -> Unet:
        """

        Args:
            input_channels: List[int]
                default value is [2048 + 1024, 128 + 512, 64 + 256].
            output_channels: List[int]
                default value is [128, 64, 32].
            is_training: bool
                default value is True.
        Returns:
            Implemented Unet.
        """

        encoder         = resnet50(is_training=is_training, catch_output=True)
        input_channels  = input_channels or [2048 + 1024, 128 + 512, 64 + 256]
        output_channels = output_channels or [128, 64, 32]

        decoder = shared_convolution_decoder(input_channels=input_channels,
                                             output_channels=output_channels,
                                             is_training=is_training)
        shared_convolution = cls(obj_name="shared_convolution",
                                 downsampling=encoder,
                                 upsampling=decoder)
        return shared_convolution




from typing import NoReturn


class DataPath:

    def __init__(self, batch_size: int, total_data: int, index_table: list):
        self._batch_size      = batch_size
        self._total_data      = total_data
        self._index_table     = index_table

        self._training_data_paths = []
        self._label_data_paths    = []

    @property
    def training_data_paths(self):
        return self._training_data_paths

    @training_data_paths.setter
    def training_data_paths(self, path: str) -> NoReturn:
        self._training_data_paths.append(path)

    @property
    def label_data_paths(self):
        return self._label_data_paths

    @label_data_paths.setter
    def label_data_paths(self, path: str) -> NoReturn:
        self._label_data_paths.append(path)

    def range(self, step: int) -> range:
        self.reset()
        start_index = self._batch_size * step + 1
        end_index   = min(self._total_data, start_index + self._batch_size)
        return self._index_table[start_index: end_index]

    def reset(self) -> NoReturn:
        self._training_data_paths.clear()
        self._label_data_paths.clear()


from typing import List, NoReturn, Tuple, Callable, Union, Optional
import tensorflow as tf
import math
from PIL import Image


class RoIRotate:

    def __init__(self, batch_size: int, target_height: int):
        self._batch_size    = batch_size
        self._target_height = target_height

    @property
    def batch_size(self) -> int:
        return self._batch_size

    @property
    def target_height(self) -> tf.Tensor:
        return tf.constant(self._target_height)

    def start(self,
              images: tf.Tensor,
              bbox_points: tf.Tensor,
              target_words: List[List[str]]
              ) -> Tuple[tf.Tensor, int, List[str], List[tf.Tensor]]:
        """
        Start roi rotate.

        Args:
        images: tf.Tensor
            Input is images or feature maps, type is tensor.
        bbox_points: tf.Tensor
            Bounding box points, format is [[x1, y1],[x2, y2],[x3, y3],[x4, y4]].
        target_words: List[List[str]]
            Target word of each text box.

        Returns:
            feature_maps: tf.Tensor
                Result of RoI rotate, feature map is text image.
            textbox_number: int
            words_stacked: List[str]
                Stacked of all target words.
            wrapped_widths: List[tf.Tensor]
                Stacked of all textbox widths.
        """

        words_stacked = []
        datasets      = []
        _ = [
            [datasets.extend(map(self.textbox_from(image=image), bounding_boxes_points)),
            words_stacked.extend(words)]
            for image, bounding_boxes_points, words in zip(images, bbox_points, target_words)
            if bounding_boxes_points != []
            ]

        wrapped_widths = [dataset[1] for dataset in datasets]  # datasets is ((textbox, width), (textbox, width), ...)
        max_width      = tf.reduce_max(tf.convert_to_tensor(wrapped_widths), axis=0)
        textbox_number = len(datasets)
        feature_maps   = tuple(map(self.padding_to(max_width=max_width), datasets))
        feature_maps   = tf.concat(feature_maps, axis=0)
        return feature_maps, textbox_number, words_stacked, wrapped_widths

    def textbox_from(self, image: tf.Tensor) -> Callable[[tf.Tensor], Tuple[tf.Tensor, tf.Tensor]]:
        """
        Preload image.

        Args:
            image: tf.Tensor

        Returns:
            affine_transformation:  Callable[[tf.Tensor], Tuple[tf.Tensor, tf.Tensor]]
        """

        image_height, image_width = image.shape[0], image.shape[1]

        def _demarcated(bbox_points: tf.Tensor) -> Tuple[tf.Tensor, tf.Tensor]:
            result = self.crop_image(image=image,
                                     bbox_points=bbox_points,
                                     image_height=image_height,
                                     image_width=image_width)

            if result is not None:
                crop_img, offset_width, offset_height, subimage_width, subimage_height = result

                zoom_rate, textbox_width  = self.calculate_resize_scale(points=bbox_points)
                target_image_coordinate   = self.create_coordinate_matrix(image_height=self.target_height, image_width=textbox_width)
                original_image_coordinate = self.create_coordinate_matrix(image_height=subimage_height, image_width=subimage_width)

                affine_matrix     = self.create_affine_matrix(zoom_rate=zoom_rate,
                                                              bbox_points=bbox_points,
                                                              left_top_point=[bbox_points[0, 1] - offset_height,
                                                                              bbox_points[0, 0] - offset_width])
                # Mapping to original coordinate.
                mapping_coordinate = tf.matmul(target_image_coordinate, tf.transpose(affine_matrix, [1, 0]))
                textbox           = self.roi_rotate(crop_img=crop_img,
                                                    mapping_coordinates=mapping_coordinate,
                                                    original_coordinate=original_image_coordinate,
                                                    textbox_width=textbox_width)
                return textbox, textbox_width
        return _demarcated

    def calculate_resize_scale(self, points: tf.Tensor):
        top_bound_length  = self.calculate_length(point_1=points[1, :], point_2=points[0, :])
        left_bound_length = self.calculate_length(point_1=points[3, :], point_2=points[0, :])

        zoom_rate     = tf.cast(self.target_height, dtype=tf.float32) / left_bound_length
        textbox_width = tf.cast(tf.math.ceil(top_bound_length * zoom_rate), dtype=tf.int32)
        return zoom_rate, textbox_width

    def create_affine_matrix(self,
                             zoom_rate: tf.float32,
                             bbox_points: tf.Tensor,
                             left_top_point: List[int]) -> tf.Tensor:
        angle  = self.calculate_angle(start_point=bbox_points[0, :], end_point=bbox_points[1, :])
        y_point, x_point = -1 * tf.cast(left_top_point, dtype=tf.float32)
        affine_matrix = zoom_rate * tf.convert_to_tensor([[tf.cos(angle), tf.sin(angle), (x_point * tf.cos(angle)) + (y_point * tf.sin(angle))],
                                                   [-tf.sin(angle), tf.cos(angle), (y_point * tf.cos(angle)) - (x_point * tf.sin(angle))],
                                                   [0, 0, 1 / zoom_rate]],
                                                  dtype=tf.float32)
        affine_matrix = tf.linalg.inv(affine_matrix)
        return affine_matrix

    def create_coordinate_matrix(self, image_height: Union[int, tf.Tensor], image_width: Union[int, tf.Tensor]) -> tf.Tensor:
        x_axes = tf.tile(tf.range(1, image_width + 1, dtype=tf.float32)[tf.newaxis, :, tf.newaxis], [image_height, 1, 1])
        y_axes = tf.tile(tf.range(1, image_height + 1, dtype=tf.float32)[:, tf.newaxis, tf.newaxis], [1, image_width, 1])
        z_axes = tf.ones(shape=[image_height, image_width, 1], dtype=tf.float32)
        return tf.concat([x_axes, y_axes, z_axes], axis=-1)

    def calculate_length(self, point_1: tf.Tensor, point_2: tf.Tensor) -> tf.Tensor:
        return tf.linalg.norm(tf.cast(point_1 - point_2, dtype=tf.float32), axis=0)

    def calculate_angle(self,
                        start_point: tf.Tensor,
                        end_point: tf.Tensor,
                        axis: Optional[str]="x") -> tf.Tensor:
        """
        Calculate the angle between the line segment and the x or y axis.

        Args:
            start_point: tf.Tensor
            end_point: tf.Tensor
            axis: Optional[str]="x"
                It is optional x or y, default is x axis.

        Returns:
            angle: tf.Tensor
                Angle of line segment.
        """
        vector = tf.cast(end_point - start_point, dtype=tf.float32)
        length = tf.linalg.norm(vector)

        x_partition, y_partition = vector
        if axis == "x":
            if x_partition != 0:
                angle = tf.acos(x_partition / length) / tf.cond(y_partition >= 0, lambda: 1, lambda: -1)
            else:
                angle = tf.cond(y_partition >= 0, lambda: math.pi / 2, lambda: -math.pi / 2)
        elif axis == "y":
            if y_partition != 0:
                angle = tf.acos(y_partition / length) / tf.cond(x_partition >= 0, lambda: 1, lambda: -1)
            else:
                angle = tf.cond(x_partition >= 0, lambda: math.pi / 2, lambda: -math.pi / 2)
        else:
            raise KeyError(f"Undefined axis {axis}, place use \"x\" or \"y\" axis.")
        return angle

    def crop_image(self,
                   image: tf.Tensor,
                   bbox_points: tf.Tensor,
                   image_height: tf.Tensor,
                   image_width: tf.Tensor
                   ) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor, tf.Tensor, tf.Tensor]:
        max_point = tf.cast(tf.reduce_max(bbox_points, axis=0), tf.int32)
        min_point = tf.cast(tf.reduce_min(bbox_points, axis=0), tf.int32)
        offset_width, offset_height = min_point[0], min_point[1]
        target_width, target_height = tf.minimum(max_point[0], image_width) - min_point[0], tf.minimum(max_point[1], image_height) - min_point[1]
        if target_width > 0 and target_height > 0:
            crop_img = tf.cast(tf.image.crop_to_bounding_box(image=image,
                                                             offset_width=offset_width,
                                                             offset_height=offset_height,
                                                             target_width=target_width,
                                                             target_height=target_height), dtype=tf.float32)
            return crop_img, offset_width, offset_height, target_width, target_height

    def roi_rotate(self,
                   crop_img: tf.Tensor,
                   mapping_coordinates: tf.Tensor,
                   original_coordinate: tf.Tensor,
                   textbox_width: tf.Tensor) -> tf.Tensor:
        value_stacked = []
        for mapping_coordinate in tf.reshape(mapping_coordinates, [-1, 3]):
            scale_rate = tf.math.maximum(0, 1 - tf.abs(original_coordinate - mapping_coordinate))
            x_scale_rate, y_scale_rate, _ = tf.split(scale_rate, 3, axis=-1)
            bilinear_interpolation = tf.reduce_sum(tf.reduce_sum(crop_img * x_scale_rate, axis=1) * y_scale_rate[:, 0], axis=0)
            value_stacked.append(bilinear_interpolation)
        del crop_img
        textbox = tf.reshape(value_stacked, [self.target_height, textbox_width, -1])
        return textbox

    def padding_to(self, max_width: tf.int32) -> Callable[[Tuple[tf.Tensor, tf.Tensor]], tf.Tensor]:
        def zero_padded(dataset: Tuple[tf.Tensor, tf.Tensor]) -> tf.Tensor:
            textbox, textbox_width = dataset

            padded_textbox = tf.pad(textbox[tf.newaxis, :, :, :], paddings=[[0, 0],
                                                                            [0, 0],
                                                                            [0, max_width - textbox_width],
                                                                            [0, 0]])
            return padded_textbox
        return zero_padded


if __name__ == "__main__":
    import numpy as np
    file_path = "/Users/abnertsai/Documents/ICDAR/ch4/training_data/img_781.jpg"
    target_path = "/Users/abnertsai/Documents/ICDAR/ch4/ch4_training_localization_transcription_gt/gt_img_781.txt"
    image = np.array(Image.open(file_path))[np.newaxis, :, :, :]
    img = Image.fromarray(image[0])
    # img.show()
    print(image.shape)
    fp = open(target_path, 'r', encoding="utf-8-sig")
    points = []
    words  = []
    for i in fp:
        points.append([int(j) for j in i.strip().split(",")[:8]])
        words.append(i.strip().split(",")[8:])
    rotate = RoIRotate(1, 16)
    images, images_number, total_words, total_width = rotate.start(images=image,
                                                                   bbox_points=[np.array(points).reshape([-1, 4, 2])],
                                                                   target_words=[words])

    for img in images:
        img = Image.fromarray(np.array(img).astype(np.uint8))
        img.show()



from ospark.models.former import Former
from ospark.nn.block import Block
from typing import List, NoReturn, Optional
import tensorflow as tf

class PositionalEncodingDelegate:

    def __init__(self):
        pass

class Informer(Former):

    def __init__(self,
                 obj_name: str,
                 encoder_blocks: List[Block],
                 class_number: int,
                 embedding_size: int,
                 dropout_rate: float,
                 positional_encoding_delegate: PositionalEncodingDelegate,
                 trained_weights: Optional[dict]=None,
                 is_training: Optional[bool]=None,
                 decoder_blocks: Optional[List[Block]]=None,
                 max_length: int = 2000,
                 ) -> NoReturn:
        super().__init__(obj_name=obj_name,
                         trained_weights=trained_weights,
                         encoder_blocks=encoder_blocks,
                         class_number=class_number,
                         embedding_size=embedding_size,
                         decoder_blocks=decoder_blocks,
                         max_length=max_length,
                         dropout_rate=dropout_rate,
                         is_training=is_training)
        self._positional_encoding = positional_encoding_delegate or PositionalEncodingDelegate()

    def pipeline(self, encoder_input: tf.Tensor, decoder_input: tf.Tensor=None) -> tf.Tensor:
        encoder_padding_mask, encoder_encoding_mask, prediction_mask = self.create_mask_matrix(encoder_input=encoder_input)
        encoder_input = self.positional_encoding(encoder_input, encoder_encoding_mask)
        encoder_input = self.encoder_dropout_layer(encoder_input, training=self.is_training)
        output = encoder_input
        for encoder_block in self.encoder_blocks:
            output = encoder_block(output)
        if self.decoder_blocks != []:
            decoder_padding_mask, decoder_encoding_mask, prediction_mask = self.create_mask_matrix(decoder_input)
            decoder_input = self.positional_encoding(decoder_input, decoder_encoding_mask)
            decoder_input = self.decoder_dropout_layer(decoder_input, training=self.is_training)
            encoder_output = output
            output = decoder_input
            for decoder_block in self.decoder_blocks:
                output = decoder_block(input_data=output, encoder_output=encoder_output)
        prediction = self.classifier(tf.matmul(output, self.classify_layer.value))
        return prediction * prediction_mask



from ospark.nn.component.basic_module import ModelObject, MetaObject
from typing import Optional, NoReturn, List
from ospark import WeightOperator
from abc import abstractmethod
import tensorflow as tf


class ModelMeta(MetaObject):

    def __call__(self, *args, **kwargs):
        obj = super().__call__(*args, **kwargs)
        obj.create()
        WeightOperator.clean_weights()
        return obj


class Model(ModelObject, metaclass=ModelMeta):

    def __init__(self,
                 obj_name: str,
                 is_training: Optional[bool]=None,
                 training_phase: Optional[bool]=None,
                 trained_weights: Optional[dict]=None):
        super(Model, self).__init__(obj_name=obj_name, is_training=is_training, training_phase=training_phase)
        self._weight_operator = WeightOperator()
        self._weight_operator.restore(weights=trained_weights or {})

    @property
    def training_weights(self) -> List[tf.Tensor]:
        weights = self._weight_operator.collect_weights(partition_name=self.obj_name)
        return weights

    def get_weights(self) -> dict:
        return self._weight_operator.weights

    def replace_weights(self, weights: dict) -> NoReturn:
        self._weight_operator.restore(weights=weights)
        self.create()
        self._weight_operator.clean_weights()

    @abstractmethod
    def pipeline(self, *args, **kwargs) -> tf.Tensor:
        pass


from __future__ import annotations
from typing import Optional, NoReturn, List, Tuple, Callable
import numpy as np
import tensorflow as tf
from . import DataGenerator
from ospark.data.encoder import LanguageDataEncoder
from tensorflow_datasets.core.deprecated.text.subword_text_encoder import SubwordTextEncoder
import math


class DataLength:

    def __init__(self):
        self._train_data_lens  = []
        self._target_data_lens = []

    @property
    def train_data_lens(self) -> list:
        return self._train_data_lens

    @property
    def target_data_lens(self) -> list:
        return self._target_data_lens

    @train_data_lens.setter
    def train_data_lens(self, datum_len: int):
        self._train_data_lens.append(datum_len)

    @target_data_lens.setter
    def target_data_lens(self, datum_len: int):
        self._target_data_lens.append(datum_len)

    @property
    def train_data_max_len(self) -> int:
        return max(self.train_data_lens)

    @property
    def target_data_max_len(self) -> int:
        return max(self.target_data_lens)

    def clear(self) -> NoReturn:
        self._train_data_lens.clear()
        self._target_data_lens.clear()


class TranslateDataGenerator(DataGenerator):

    class Dataset:

        def __init__(self):
            self._training_data = []
            self._target_data   = []
            self._data_lens     = DataLength()

        @property
        def training_data(self) -> tf.Tensor:
            return tf.concat([tf.pad(tf.convert_to_tensor(datum)[tf.newaxis, :], [[0, 0], [0, self.length.train_data_max_len - datum_len]]) for datum, datum_len
                              in zip(self._training_data, self.length.train_data_lens)], axis=0)

        @property
        def target_data(self) -> tf.Tensor:
            return tf.concat([tf.pad(tf.convert_to_tensor(datum)[tf.newaxis, :], [[0, 0], [0, self.length.target_data_max_len - datum_len]]) for datum, datum_len
                              in zip(self._target_data, self.length.target_data_lens)], axis=0)

        @training_data.setter
        def training_data(self, datum: tf.Tensor):
            self._training_data.append(datum)
            self._data_lens.train_data_lens = len(datum)

        @target_data.setter
        def target_data(self, datum: tf.Tensor):
            self._target_data.append(datum)
            self._data_lens.target_data_lens = len(datum)

        @property
        def batch_number(self) -> int:
            return len(self._training_data)

        @property
        def length(self) -> DataLength:
            return self._data_lens

        def clear(self) -> NoReturn:
            self._training_data.clear()
            self._target_data.clear()
            self._data_lens.clear()

    def __init__(self,
                 training_data: List[str],
                 target_data: List[str],
                 data_encoder: LanguageDataEncoder,
                 batch_size: int,
                 max_token: Optional[int]=None,
                 max_length: Optional[int]=None,
                 start_index: Optional[int]=None,
                 padding_range: Optional[int]=None) -> NoReturn:
        super().__init__(training_data=training_data, target_data=target_data, batch_size=batch_size, initial_step=0)
        self._data_number         = len(training_data)
        self._data_encoder        = data_encoder
        self._train_data_bos      = [data_encoder.train_data_encoder.vocab_size]  # 因為 vocab 中並沒有 bos 及 eos，所以新增額外兩個 index 當作 bos 及 eos
        self._target_data_bos     = [data_encoder.label_data_encoder.vocab_size]
        self._train_data_eos      = [data_encoder.train_data_encoder.vocab_size + 1]
        self._target_data_eos     = [data_encoder.label_data_encoder.vocab_size + 1]
        self._padding_range       = padding_range or 50
        self._max_length          = max_length
        self._next_interval       = None
        self._max_token           = max_token or 3000
        self._element_index       = start_index or 0
        self._get_data            = self.from_batch_limit if max_token is None else self.from_token_limit

    @property
    def data_number(self) -> int:
        return self._data_number

    @property
    def data_encoder(self) -> LanguageDataEncoder:
        return self._data_encoder

    @property
    def train_data_bos(self) -> List[int]:
        return self._train_data_bos

    @property
    def target_data_bos(self) -> List[int]:
        return self._target_data_bos

    @property
    def train_data_eos(self) -> List[int]:
        return self._train_data_eos

    @property
    def target_data_eos(self) -> List[int]:
        return self._target_data_eos

    @property
    def max_length(self) -> int:
        return self._max_length

    @property
    def padding_range(self) -> int:
        return self._padding_range

    @property
    def max_token(self) -> int:
        return self._max_token

    @property
    def element_index(self) -> int:
        return self._element_index

    def encode_bos_eos(self,
                       train_sequence: str,
                       target_sequence: str) -> Tuple[np.ndarray, np.ndarray]:
        if isinstance(train_sequence, tf.Tensor):
            train_sequence  = train_sequence.numpy().decode("utf-8")
            target_sequence = target_sequence.numpy().decode("utf-8")

        train_sequence = self.train_data_bos + self.data_encoder.encode_train_data(train_sequence) + self.train_data_eos
        target_sequence = self.target_data_bos + self.data_encoder.encode_label_data(target_sequence) + self.target_data_eos
        return np.array(train_sequence), np.array(target_sequence)

    def filter_length(self, train_sequence: np.ndarray, target_sequence: np.ndarray) -> bool:
        return len(train_sequence) > self.max_length or len(target_sequence) > self.max_length

    def from_batch_limit(self) -> Dataset:
        self.dataset.clear()
        while self.element_index < self.data_number:
            train_sequence, target_sequence = self.encode_bos_eos(train_sequence=self.training_data[self.element_index],
                                                                  target_sequence=self.target_data[self.element_index])
            if self.max_length is not None:
                is_filtered = self.filter_length(train_sequence=train_sequence, target_sequence=target_sequence)
                if is_filtered:
                    self._element_index += 1
                    continue
            self.dataset.training_data = train_sequence
            self.dataset.target_data   = target_sequence
            self._element_index += 1

            if self.batch_size == self.dataset.batch_number:
                break

        return self.dataset

    def from_token_limit(self) -> Dataset:
        padding_length = 0
        self.dataset.clear()
        while self.element_index < self.data_number:
            train_sequence, target_sequence = self.encode_bos_eos(train_sequence=self.training_data[self.element_index],
                                                                  target_sequence=self.target_data[self.element_index])
            if self.max_length is not None:
                is_filtered = self.filter_length(train_sequence=train_sequence, target_sequence=target_sequence)
                if is_filtered:
                    self._element_index += 1
                    continue

            self.dataset.training_data = train_sequence
            self.dataset.target_data   = target_sequence
            if self.padding_range is not None:
                padding_length = max(math.ceil(max(self.dataset.length.train_data_max_len, self.dataset.length.target_data_max_len) / self.padding_range) * self.padding_range,
                                     padding_length)
            else:
                padding_length = max(max(self.dataset.length.train_data_max_len, self.dataset.length.target_data_max_len), padding_length)

            token_number = padding_length * self.dataset.batch_number

            if self.max_token < token_number:
                self.dataset.length.train_data_lens.pop()
                self.dataset.length.target_data_lens.pop()
                self.dataset._training_data.pop()
                self.dataset._target_data.pop()
                break
            else:
                self._element_index += 1

        return self.dataset

    def __next__(self) -> Dataset:
        if self.element_index < self.data_number:
            return self._get_data()
        self.reset()
        raise StopIteration()

    def reset(self) -> NoReturn:
        self._element_index = 0

    def __call__(self):
        return self

    @classmethod
    def transform_into_tf_datasets(cls,
                                   train_data: List[str],
                                   target_data: List[str],
                                   data_encoder: LanguageDataEncoder,
                                   batch_size: int,
                                   max_token: Optional[int]=None,
                                   max_length: Optional[int]=None,
                                   start_index: Optional[int]=None,
                                   padding_range: Optional[int]=None) -> tf.data.Dataset:
        generator = cls(training_data=train_data,
                        target_data=target_data,
                        data_encoder=data_encoder,
                        batch_size=batch_size,
                        max_token=max_token,
                        max_length=max_length,
                        start_index=start_index,
                        padding_range=padding_range)
        generator = tf.data.Dataset.from_generator(generator,
                                                   output_signature=(tf.TensorSpec(shape=(None, None), dtype=tf.int64),
                                                                     tf.TensorSpec(shape=(None, None), dtype=tf.int64)))
        return generator


from ospark.nn.block import Block
from ospark.nn.layers import Layer
from ospark.nn.layers.connection_layer import ConnectionLayer
from typing import List, NoReturn
from functools import reduce
import tensorflow as tf


class ConnectionBlock(Block):

    def __init__(self,
                 obj_name: str,
                 connection_layers: List[ConnectionLayer]):
        super().__init__(obj_name=obj_name)
        self._connection_layers = connection_layers

    @property
    def connection_layers(self) -> List[Layer]:
        return self._connection_layers

    def in_creating(self) -> NoReturn:
        for layer in self.connection_layers:
            self.assign(component=layer)

    def pipeline(self, input_data: tf.Tensor, connection_input: List[tf.Tensor]) -> tf.Tensor:
        """

        : input_data is 4-D Tensor [b, h, w, c]:
        :param connection_input:
        :return:
        """

        return reduce(
            lambda output, layer: layer.pipeline(output, connection_input.pop()),
            self.connection_layers,
            input_data)


def shared_convolution_decoder(input_channels: List[int],
                               output_channels: List[int],
                               is_training: bool) -> ConnectionBlock:
    connection_layers = []
    for i, channels in enumerate(zip(input_channels, output_channels)):
        input_channel, output_channel = channels
        name = f"connection_layer_{i}"
        connection_layer = ConnectionLayer(obj_name=name,
                                           concatenated_channel=input_channel,
                                           output_channel=output_channel,
                                           is_training=is_training)
        connection_layers.append(connection_layer)
    decoder = ConnectionBlock(obj_name="shared_decoder",
                              connection_layers=connection_layers)
    return decoder





from ospark.models.transformer import *
from ospark.nn.block.deep_transformer_block import DeepTransformerEncoder, DeepTransformerDecoder


class ExdeepTransformer(Transformer):

    def __init__(self,
                 obj_name: str,
                 encoder_blocks: List[DeepTransformerEncoder],
                 class_number: int,
                 embedding_size: int,
                 dropout_rate: float,
                 trained_weights: Optional[dict]=None,
                 is_training: Optional[bool]=False,
                 encoder_corpus_size: Optional[int]=None,
                 decoder_corpus_size: Optional[int]=None,
                 use_embedding_layer: Optional[bool]=True,
                 use_classifier: Optional[bool]=False,
                 decoder_blocks: Optional[List[DeepTransformerDecoder]]=None,
                 max_length: int=2000,
                 ) -> NoReturn:
        super().__init__(obj_name=obj_name,
                         trained_weights=trained_weights,
                         encoder_blocks=encoder_blocks,
                         class_number=class_number,
                         embedding_size=embedding_size,
                         use_embedding_layer=use_embedding_layer,
                         encoder_corpus_size=encoder_corpus_size,
                         decoder_corpus_size=decoder_corpus_size,
                         use_classifier=use_classifier,
                         decoder_blocks=decoder_blocks,
                         max_length=max_length,
                         dropout_rate=dropout_rate,
                         is_training=is_training
                         )

    @property
    def encoder_blocks(self) -> List[DeepTransformerEncoder]:
        return self._encoder_blocks

    @property
    def decoder_blocks(self) -> List[DeepTransformerDecoder]:
        return self._decoder_blocks


    def profiling_phase(self, encoder_input: tf.Tensor, decoder_input) -> NoReturn:
        catch_list = []
        seq_len        = tf.shape(encoder_input)[1]
        encoder_padding_mask, lookahead_mask = self.create_mask_matrix(encoder_input, decoder_input)

        encoder_input  = self.encoder_embedding_layer.pipeline(encoder_input)
        encoder_input *= self.embedding_scale_rate
        encoder_input += self.positional_encoding_table[:, seq_len, :]
        encoder_input  = self.encoder_dropout_layer(encoder_input, training=self.is_training)
        output = encoder_input
        for encoder_block in self.encoder_blocks:
            output, catch_list = encoder_block.profiling_phase(input_data=output,
                                                               catch_variance=catch_list,
                                                               mask=encoder_padding_mask)

        if self.decoder_blocks != []:
            seq_len        = tf.shape(encoder_input)[1]
            decoder_input  = self.decoder_embedding_layer(decoder_input)
            decoder_input *= self.embedding_scale_rate
            decoder_input += self.positional_encoding_table[:, seq_len, :]
            decoder_input  = self.decoder_dropout_layer(decoder_input, training=self.is_training)
            encoder_output = output
            output = decoder_input
            for decoder_block in self.decoder_blocks:
                output, catch_list = decoder_block.profiling_phase(input_data=output,
                                                                   encoder_output=encoder_output,
                                                                   catch_variance=catch_list,
                                                                   encoder_padding_mask=encoder_padding_mask,
                                                                   decoder_padding_mask=lookahead_mask)

    def back_to_standard(self) -> NoReturn:
        _ = [encoder_block.recover_transformer() for encoder_block in self.encoder_blocks]
        if self.decoder_blocks != []:
            _ = [decoder_block.recover_transformer() for decoder_block in self.decoder_blocks]



import tensorflow as tf

from ospark.nn.loss_function.loss_function import LossFunction


class Focal(LossFunction):

    def __init__(self, alpha: float, beta: float):
        self._alpha  = alpha
        self._beta   = beta

    @property
    def alpha(self) -> float:
        return self._alpha

    @property
    def beta(self) -> float:
        return self._beta

    def calculate_loss(self, prediction: tf.Tensor, target_data: tf.Tensor) -> tf.Tensor:
        A_part = self.alpha * target_data * tf.math.pow(1 - prediction, self.gamma) * tf.math.log(prediction)
        B_part = (1 - self.alpha) * (1 - target_data) * tf.math.pow(prediction, self.gamma) * tf.math.log(1 - prediction)
        loss = -tf.reduce_mean(A_part + B_part)
        return loss


import tensorflow as tf

import ospark.utility.weight_initializer
from ospark.nn.layers import Layer
from ospark.nn.layers.normalization import Normalization
from typing import *
import ospark
import time


class SelfAttentionLayer(Layer):

    def __init__(self, 
                 obj_name: str, 
                 embedding_size: int, 
                 head_number: int,
                 dropout_rate: float,
                 training_phase: Optional[bool]=None,
                 is_training: Optional[bool]=None,
                 use_look_ahead: Optional[bool]=None) -> NoReturn:
        super().__init__(obj_name=obj_name, is_training=is_training, training_phase=training_phase)
        assert embedding_size % head_number == 0

        self._depth           = embedding_size // head_number
        self._embedding_size  = embedding_size
        self._head_number     = head_number
        self._sequence_length = None
        self._dropout_layer   = tf.keras.layers.Dropout(rate=dropout_rate)
        self._use_look_ahead  = use_look_ahead or False

    @property
    def depth(self) -> int:
        return self._depth

    @property
    def embedding_size(self) -> int:
        return self._embedding_size

    @property
    def head_number(self) -> int:
        return self._head_number

    @property
    def sequence_length(self) -> None:
        return self._sequence_length

    @property
    def dropout_layer(self) -> tf.keras.layers.Dropout:
        return self._dropout_layer

    @property
    def use_look_ahead(self) -> bool:
        return self._use_look_ahead

    @property
    def look_ahead_mask(self) -> tf.Tensor:
        return 1 - tf.linalg.band_part(tf.ones((self.sequence_length, self.sequence_length)), -1, 0)

    def in_creating(self) -> NoReturn:
        self._q_weights = ospark.utility.weight_initializer.glorot_uniform(
            obj_name="q_weights",
            shape=[self.embedding_size, self.embedding_size]
        )
        self._k_weights = ospark.utility.weight_initializer.glorot_uniform(
            obj_name="k_weights",
            shape=[self.embedding_size, self.embedding_size]
        )
        self._v_weights = ospark.utility.weight_initializer.glorot_uniform(
            obj_name="v_weights",
            shape=[self.embedding_size, self.embedding_size]
        )
        self._q_bias = ospark.utility.weight_initializer.zeros(
            obj_name="q_bias",
            shape=[self.embedding_size]
        )
        self._k_bias = ospark.utility.weight_initializer.zeros(
            obj_name="k_bias",
            shape=[self.embedding_size]
        )
        self._v_bias = ospark.utility.weight_initializer.zeros(
            obj_name="v_bias",
            shape=[self.embedding_size]
        )
        self._output_weights = ospark.utility.weight_initializer.glorot_uniform(
            obj_name="output_weights",
            shape=[self.embedding_size, self.embedding_size]
        )
        self._output_bias = ospark.utility.weight_initializer.zeros(
            obj_name="output_bias",
            shape=[self.embedding_size]
        )
        self._norm = ospark.nn.layers.normalization.LayerNormalization(obj_name="layer_norm",
                                                                       layer_dimension=self.embedding_size)

    def pipeline(self, input_data: tf.Tensor, mask: Optional[tf.Tensor]=None, *args, **kwargs):
        return self.layer_calculation(Q_input=input_data,
                                      K_input=input_data,
                                      V_input=input_data,
                                      mask=mask)

    def layer_calculation(self,
                          Q_input: tf.Tensor,
                          K_input: tf.Tensor,
                          V_input: tf.Tensor,
                          mask: Optional[tf.Tensor] = None) -> tf.Tensor:
        batch_size = tf.shape(Q_input)[0]

        Q, K, V     = self.QKV_process(Q_input=Q_input, K_input=K_input, V_input=V_input, batch_size=batch_size)
        main_output = self.attention_layer(Q=Q, K=K, V=V, batch_size=batch_size, mask=mask)

        residual_output = self.residual_net(input_data=Q_input)
        added_residual  = tf.add(self.dropout_layer(main_output, training=self.training_phase), residual_output)
        layer_output    = self._norm(added_residual)
        return layer_output

    def QKV_process(self,
                    Q_input: tf.Tensor,
                    K_input: tf.Tensor,
                    V_input: tf.Tensor,
                    batch_size: tf.Tensor) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
        Q = tf.matmul(Q_input, self._q_weights) + self._q_bias  # [batch, seq, d_model]
        K = tf.matmul(K_input, self._k_weights) + self._k_bias
        V = tf.matmul(V_input, self._v_weights) + self._v_bias
        Q = self.split_head(input_data=Q, batch_size=batch_size) # [batch, head_number, seq, depth]
        K = self.split_head(input_data=K, batch_size=batch_size)
        V = self.split_head(input_data=V, batch_size=batch_size)
        return Q, K, V

    def split_head(self, input_data: tf.Tensor, batch_size: tf.int32) -> tf.Tensor:
        split_result     = tf.reshape(input_data, [batch_size, -1, self.head_number, self.depth]) # [batch, seq, head_number, depth]
        transpose_result = tf.transpose(split_result, [0, 2, 1, 3]) # [batch, head_number, seq, depth]
        return transpose_result

    def attention_layer(self, 
                        Q: tf.Tensor, 
                        K: tf.Tensor, 
                        V: tf.Tensor,
                        batch_size: tf.Tensor,
                        mask: Optional[tf.Tensor]=None) -> tf.Tensor:
        attention_value = self.attention(Q=Q, K=K, V=V, batch_size=batch_size, mask=mask)
        layer_output    = tf.matmul(attention_value, self._output_weights) + self._output_bias
        return layer_output

    def attention(self,
                  Q: tf.Tensor,
                  K: tf.Tensor,
                  V: tf.Tensor,
                  batch_size: tf.Tensor,
                  mask: Optional[tf.Tensor]=None) -> tf.Tensor:
        K = tf.transpose(K, [0, 1, 3, 2])  # BHLD -> BHDL
        scaled_dot_product = tf.matmul(Q, K) / tf.math.sqrt(tf.cast(self.embedding_size, dtype=tf.float32))  # BHLD * BHDL -> BHLL
        if self.use_look_ahead and mask is not None:
            self._sequence_length = tf.shape(Q)[-2]

            mask = (tf.cast(tf.math.not_equal(mask + self.look_ahead_mask, 0), tf.float32) * -1e9)[:, tf.newaxis, ...]

            scaled_dot_product += mask
        elif mask is not None:
            scaled_dot_product += (mask * -1e9)
        scaled_dot_product = tf.nn.softmax(scaled_dot_product, axis=-1)
        scaled_attention   = tf.matmul(scaled_dot_product, V) # BHLL * BHLD -> BHLD
        scaled_attention   = tf.transpose(scaled_attention, [0, 2, 1, 3]) # BHLD -> BLHD
        concat_output      = tf.reshape(scaled_attention, [batch_size, -1, self.embedding_size])
        return concat_output

    def residual_net(self, input_data: tf.Tensor) -> tf.Tensor:
        return input_data


class EncoderDecoderAttentionLayer(SelfAttentionLayer):

    def __init__(self,
                 obj_name: str, 
                 embedding_size: int, 
                 head_number: int,
                 dropout_rate: float,
                 is_training: Optional[bool]=None,
                 use_look_ahead: Optional[bool]=None) -> NoReturn:
        super().__init__(obj_name=obj_name, 
                         embedding_size=embedding_size, 
                         head_number=head_number,
                         dropout_rate=dropout_rate,
                         is_training=is_training,
                         use_look_ahead=use_look_ahead)

    def pipeline(self, input_data: tf.Tensor, encoder_output: tf.Tensor, mask: Optional[tf.Tensor]=None):
        return self.layer_calculation(Q_input=input_data,
                                      K_input=encoder_output,
                                      V_input=encoder_output,
                                      mask=mask)


class KernelFunction:

    def __init__(self):
        pass

    def __call__(self, *args, **kwargs):
        raise NotImplementedError()


class SoftmaxKernel(KernelFunction):

    def __call__(self, input_data: tf.Tensor, projection_matrix: tf.Tensor):
        normalizer = 1.0 / tf.math.sqrt(tf.cast(input_data.shape[-1], dtype=tf.float32))
        input_data = input_data * normalizer
        scaling    = 1.0 / tf.math.sqrt(tf.cast(projection_matrix.shape[0], dtype=tf.float32))

        mapping_result  = tf.matmul(input_data, tf.transpose(projection_matrix, [1, 0])) # blhd * dm -> blhm
        correction_part = tf.square(input_data)
        correction_part = tf.math.reduce_sum(correction_part, axis=-1, keepdims=True) / 2.0 # blh1
        result = scaling * tf.math.exp(mapping_result - correction_part)
        return result


@tf.custom_gradient
def causal_numerator(qs, ks, vs):
  """Computes not-normalized FAVOR causal attention A_{masked}V.
  Args:
    qs: query_prime tensor of the shape [L,B,H,M].
    ks: key_prime tensor of the shape [L,B,H,M].
    vs: value tensor of the shape [L,B,H,D].
  Returns:
    Not-normalized FAVOR causal attention A_{masked}V.
  """

  result = []
  sums = tf.zeros_like(tf.einsum("ijk,ijl->ijkl", ks[0], vs[0]))

  for index in range(qs.shape[0]):
    sums = sums + tf.einsum("ijk,ijl->ijkl", ks[index], vs[index]) # BHMD
    result.append(tf.einsum("ijkl,ijk->ijl", sums, qs[index])[None, Ellipsis])

  result = tf.concat(result, axis=0)

  def grad(res_grad):

    grads = tf.zeros_like(tf.einsum("ijk,ijl->ijkl", ks[0], vs[0]))

    gr_sums = sums

    q_grads = []
    k_grads = []
    v_grads = []

    for index in range(qs.shape[0] - 1, -1, -1):

      q_grads.append(
          tf.einsum("ijkl,ijl->ijk", gr_sums, res_grad[index])[None, Ellipsis])
      grads = grads + tf.einsum("ijk,ijl->ijkl", qs[index], res_grad[index])
      k_grads.append(tf.einsum("ijkl,ijl->ijk", grads, vs[index])[None, Ellipsis])
      v_grads.append(tf.einsum("ijkl,ijk->ijl", grads, ks[index])[None, Ellipsis])
      gr_sums = gr_sums - tf.einsum("ijk,ijl->ijkl", ks[index], vs[index])

    q_grads = tf.concat(q_grads[::-1], axis=0)
    k_grads = tf.concat(k_grads[::-1], axis=0)
    v_grads = tf.concat(v_grads[::-1], axis=0)

    return q_grads, k_grads, v_grads

  return result, grad


@tf.custom_gradient
def causal_denominator(qs, ks):
  """Computes FAVOR normalizer in causal attention.
  Args:
    qs: query_prime tensor of the shape [L,B,H,M].
    ks: key_prime tensor of the shape [L,B,H,M].
  Returns:
    FAVOR normalizer in causal attention.
  """

  result = []
  sums = tf.zeros_like(ks[0])

  for index in range(qs.shape[0]):
    sums = sums + ks[index]
    result.append(tf.reduce_sum(qs[index] * sums, axis=2)[None, Ellipsis])
    # result.append(tf.einsum("ijk,ijk->ij", qs[index], sums)[None, Ellipsis])

  result = tf.concat(result, axis=0)

  def grad(res_grad):

    k_grad = tf.zeros_like(ks[0])

    gr_sums = sums

    q_grads = []
    k_grads = []

    for index in range(qs.shape[0] - 1, -1, -1):

      q_grads.append(
          tf.einsum("ijk,ij->ijk", gr_sums, res_grad[index])[None, Ellipsis])
      k_grad = k_grad + tf.einsum("ijk,ij->ijk", qs[index], res_grad[index])
      k_grads.append(k_grad[None, Ellipsis])
      gr_sums = gr_sums - ks[index]

    q_grads = tf.concat(q_grads[::-1], axis=0)
    k_grads = tf.concat(k_grads[::-1], axis=0)

    return q_grads, k_grads

  return result, grad


class FavorAttentionLayer(SelfAttentionLayer):

    def __init__(self,
                 obj_name: str,
                 embedding_size: int,
                 head_number: int,
                 dropout_rate: float,
                 random_projections_number: int,
                 kernel_function: Optional[KernelFunction]=None,
                 is_training: Optional[bool]=False,
                 use_look_ahead: Optional[bool]=False) -> NoReturn:
        super().__init__(obj_name=obj_name,
                         is_training=is_training,
                         embedding_size=embedding_size,
                         head_number=head_number,
                         dropout_rate=dropout_rate,
                         use_look_ahead=use_look_ahead)
        self._random_projections_number = random_projections_number
        self._inner_dimension           = int(embedding_size / head_number)
        self._kernel_function           = kernel_function or SoftmaxKernel()

    @property
    def random_projections_number(self) -> int:
        return self._random_projections_number

    @property
    def inner_dimension(self) -> int:
        return self._inner_dimension

    @property
    def kernel_function(self) -> KernelFunction:
        return self._kernel_function

    def create_projections_matrix(self,
                                  random_projections_number: int,
                                  embedding_size: int,
                                  seed: Optional[int]=0):
        blocks = []

        block_number = int(random_projections_number / embedding_size)
        current_seed = seed

        for _ in range(block_number):
            orthogonal_matrix = self.get_orthogonal_matrix(matrix_shape=[embedding_size, embedding_size],
                                                           seed=current_seed)
            blocks.append(orthogonal_matrix)
            current_seed = current_seed + 1

        remaining_row = random_projections_number - block_number * embedding_size
        if remaining_row > 0:
            orthogonal_matrix = self.get_orthogonal_matrix(matrix_shape=[embedding_size, embedding_size],
                                                           seed=current_seed)
            blocks.append(orthogonal_matrix[:remaining_row])

        final_matrix = tf.experimental.numpy.vstack(blocks)
        multiplier   = tf.sqrt(float(embedding_size)) * tf.ones(shape=[random_projections_number])
        return tf.linalg.matmul(tf.linalg.diag(multiplier), final_matrix)

    def get_orthogonal_matrix(self, matrix_shape: List[int], seed: int) -> tf.Tensor:
        tf.random.set_seed(seed=seed)
        random_feature = tf.random.normal(shape=matrix_shape)
        q, _ = tf.linalg.qr(random_feature)
        return q

    def attention(self,
                  Q: tf.Tensor,
                  K: tf.Tensor,
                  V: tf.Tensor,
                  batch_size: tf.Tensor,
                  mask: Optional[tf.Tensor]=None) -> tf.Tensor:
        seed = tf.cast(tf.math.ceil(tf.math.abs(tf.reduce_sum(Q) * 1e8)), tf.int32)
        projection_matrix = self.create_projections_matrix(random_projections_number=self.random_projections_number,
                                                           embedding_size=self.inner_dimension,
                                                           seed=seed)
        Q = tf.transpose(Q, [2, 0, 1, 3])  # BHLD -> LBHD
        K = tf.transpose(K, [2, 0, 1, 3])  # BHLD -> LBHD
        V = tf.transpose(V, [2, 0, 1, 3])  # BHLD -> LBHD
        Q = self.kernel_function(input_data=Q, projection_matrix=projection_matrix)  # LBHD -> LBHM
        K = self.kernel_function(input_data=K, projection_matrix=projection_matrix)  # LBHD -> LBHM
        if self.use_look_ahead:
            normalizer       = causal_numerator(qs=Q, ks=K, vs=V) # LBHD
            denominator      = causal_denominator(qs=Q, ks=K) # LBH
            attention_output = normalizer / denominator[..., tf.newaxis] # LBHD / LBH1
            attention_output = tf.transpose(attention_output, [1, 0, 2, 3]) # LBHD -> BLHD
        else:
            normalizer       = self.calculate_numerator(qs=Q, ks=K, vs=V)  # LBHD
            denominator      = self.calculate_denominator(qs=Q, ks=K)[..., tf.newaxis]  # LBH -> LBH1
            attention_output = tf.transpose(normalizer / denominator, [1, 0, 2, 3]) # LBHD -> BLHD
        attention_output     = tf.reshape(attention_output,
                                          shape=[batch_size, -1, self.embedding_size])
        return attention_output

    def calculate_numerator(self, qs: tf.Tensor, ks: tf.Tensor, vs: tf.Tensor) -> tf.Tensor:
        kvs = tf.einsum("lbhm,lbhd->bhmd", ks, vs)
        return tf.einsum("lbhm,bhmd->lbhd", qs, kvs)

    def calculate_denominator(self, qs: tf.Tensor, ks: tf.Tensor) -> tf.Tensor:
        # qs -> LBHM, ks -> LBHM
        all_ones = tf.ones([ks.shape[0]])
        ks_sum = tf.einsum("lbhm,l->bhm", ks, all_ones)
        return tf.einsum("lbhm,bhm->lbh", qs, ks_sum)

    def bidirectional_attention(self, q: tf.Tensor, k: tf.Tensor, v: tf.Tensor):
        B, L, H, D = tf.shape(v)
        c       = tf.concat([v, tf.ones(shape=[B, L, H, 1])], axis=-1) # BLHD -> BLH D+1
        # k       = tf.transpose(k, [0, 1, 3, 2]) # BLHM -> BLMH
        buf1    = tf.matmul(k, c, transpose_a=True) # BLMH * BLHD+1 -> BLMD+1
        buf2    = tf.matmul(q, buf1) # BLHM * BLMD+1 -> BLHD+1
        buf3    = buf2[:, :, :, :D] # BLHD
        buf4    = buf2[:, :, :, D:] # BLH1
        return buf3 / tf.reduce_sum(buf4, axis=1, keepdims=True) # BLHD

    def unidirectional_attention(self, q: tf.Tensor, k: tf.Tensor, v: tf.Tensor):
        B, L, H, D = tf.shape(v)

        c = tf.concat([v, tf.ones(shape=[B, L, H, 1])], axis=-1) # BLHD -> BLH D+1

        G = tf.matmul(k[..., tf.newaxis], tf.reshape(c, shape=[B, L, H, 1, D + 1])) # BLHM1 * BLH1 D+1 -> BLHM D+1


        mask = tf.reshape(1 - tf.linalg.band_part(tf.ones((L, L)), -1, 0), shape=[1, L, L , 1, 1, 1])
        mask = tf.tile(mask, [B, 1, 1, H, self.random_projections_number, D+1]) # BLLHM D+1
        s = time.time()
        Gps = tf.linalg.einsum("ijklmn, ijlmn->ijlmn", mask, G)
        print("1",time.time() - s)
        s = time.time()
        Gps = tf.concat([tf.reduce_sum(G[:, :i, ...], axis=1, keepdims=True)for i in range(L)], axis=1) # BLHM D+1
        print("2",time.time() - s)

        q = tf.reshape(q, shape=[B, L, H, 1, -1]) # BLH1M

        buf2 = tf.squeeze(tf.matmul(q, Gps))  # BLH1M * BLHM D+1 -> BLH1D+1 -> BLHD+1


        # buf2 = tf.concat([tf.reshape(tf.matmul(tf.reshape(q[:, i, ...], shape=[B, H, 1, -1]), tf.reduce_sum(G[:, :i, ...], axis=1)), [B, 1, H, D + 1])
        #        for i in range(L)], axis=1) # B1H1M * B1HM D+1 -> B1H1 D+1 for L -> BLH D+1

        buf3 = buf2[:, :, :, :D] # BLHD
        buf4 = buf2[:, :, :, D:] # BLH1
        return buf3 / tf.reduce_sum(buf4, axis=1, keepdims=True) # BLHD


class EncodeDecodeFavorAttention(FavorAttentionLayer):

    def __init__(self,
                 obj_name: str,
                 embedding_size: int,
                 head_number: int,
                 dropout_rate: float,
                 random_projections_number: int,
                 kernel_function: Optional[KernelFunction]=None,
                 is_training: Optional[bool]=False,
                 use_look_ahead: Optional[bool]=False) -> NoReturn:
        super(EncodeDecodeFavorAttention, self).__init__(obj_name=obj_name,
                                                         embedding_size=embedding_size,
                                                         head_number=head_number,
                                                         dropout_rate=dropout_rate,
                                                         random_projections_number=random_projections_number,
                                                         kernel_function=kernel_function,
                                                         is_training=is_training,
                                                         use_look_ahead=use_look_ahead)

    def pipeline(self, input_data: tf.Tensor, encoder_output: tf.Tensor, mask: Optional[tf.Tensor]=None):
        return self.layer_calculation(Q_input=input_data,
                                      K_input=encoder_output,
                                      V_input=encoder_output,
                                      mask=mask)


from __future__ import annotations
import tensorflow as tf
from ospark.nn.layers import Layer
from typing import Optional
from typing import NoReturn


class Activation(Layer):

    def __init__(self, is_training: Optional[bool]=None):
        super(Activation, self).__init__(obj_name=self.__class__.__name__, is_training=is_training)

    def __init_subclass__(cls) -> NoReturn:
        super().__init_subclass__()
        setattr(Activation, cls.__name__, cls)
    
    def pipeline(self, input_data: tf.Tensor) -> tf.Tensor:
        raise NotImplementedError()

    def __call__(self, input_data: tf.Tensor) -> tf.Tensor:
        return self.pipeline(input_data=input_data)


class PassActivation(Activation):

    def pipeline(self, input_data: tf.Tensor) -> tf.Tensor:
        return input_data


class ReLU(Activation):

    def pipeline(self, input_data: tf.Tensor) -> tf.Tensor:
        return tf.nn.relu(input_data)


class ELU(Activation):

    def pipeline(self, input_data: tf.Tensor) -> tf.Tensor:
        return tf.nn.elu(input_data)


class LeakyReLU(Activation):

    @property
    def alpha(self) -> tf.float32:
        return self.__alpha

    def __init__(self, alpha: tf.float32) -> NoReturn:
        self.__alpha = alpha
        super().__init__()

    def pipeline(self, input_data: tf.Tensor) -> tf.Tensor:
        return tf.nn.leaky_relu(input_data, self.alpha)


class SELU(Activation):

    def pipeline(self, input_data: tf.Tensor) -> tf.Tensor:
        return tf.nn.selu(input_data)


class CReLU(Activation):

    def pipeline(self, input_data: tf.Tensor) -> tf.Tensor:
        return tf.nn.crelu(input_data)


class GELU(Activation):

    def pipeline(self, input_data: tf.Tensor) -> tf.Tensor:
        return tf.nn.gelu(input_data)


class Swish(Activation):

    def __init__(self, beta: float):
        super(Swish, self).__init__()
        self._beta = tf.constant(beta)

    def pipeline(self, input_data: tf.Tensor) -> tf.Tensor:
        return tf.nn.swish(input_data, beta=self._beta)


class Sigmoid(Activation):

    def pipeline(self, input_data: tf.Tensor) -> tf.Tensor:
        return tf.nn.sigmoid(input_data)


from __future__ import annotations

import ospark.utility.weight_initializer
from ospark.nn.layers import Layer
from ospark.nn.layers.activation import Activation, ReLU, PassActivation
from ospark.nn.layers.normalization import Normalization, BatchNormalization, PassNormalization
from typing import List, NoReturn, Optional
import tensorflow as tf
import ospark


class ConvolutionLayer(Layer):

    def __init__(self,
                 obj_name: str,
                 filter_size: List[int],
                 strides: List[int],
                 padding: str,
                 is_training: Optional[bool]=None,
                 activation: Optional[Activation]=None,
                 normalization: Optional[Normalization]=None,
                 layer_order: Optional[List[str]]=None) -> NoReturn:
        super().__init__(obj_name=obj_name, is_training=is_training)
        self._filter_size   = filter_size
        self._strides       = strides
        self._padding       = padding
        self._activation    = activation or PassActivation()
        self._normalization = normalization or PassNormalization()
        self._layer_order   = layer_order or ["conv", "norm", "activate"]

    @property
    def filter_size(self) -> List[int]:
        return self._filter_size

    @property
    def strides(self) -> List[int]:
        return self._strides

    @property
    def padding(self) -> str:
        return self._padding

    @property
    def activation(self) -> Activation:
        return self._activation

    @property
    def normalization(self) -> Normalization:
        return self._normalization

    @property
    def layer_order(self) -> List[str]:
        return self._layer_order

    def set_norm(self, norm: Normalization) -> NoReturn:
        self._normalization = norm

    def set_activate(self, activation: Activation) -> NoReturn:
        self._activation = activation

    def set_order(self, layer_order: List[str]) -> NoReturn:
        self._layer_order = layer_order

    @classmethod
    def bn_relu_conv(cls,
                     obj_name: str,
                     filter_size: List[int],
                     strides: List[int],
                     padding: str,
                     is_training: bool) -> ConvolutionLayer:
        activation    = ReLU()
        normalization = BatchNormalization(obj_name="batch_norm",
                                           input_depth=filter_size[-2],
                                           is_training=is_training)
        return cls(obj_name=obj_name,
                   filter_size=filter_size,
                   strides=strides,
                   padding=padding,
                   activation=activation,
                   normalization=normalization,
                   layer_order=["norm", "activate", "conv"])

    @classmethod
    def conv_bn_relu(cls,
                     obj_name: str,
                     filter_size: List[int],
                     strides: List[int],
                     padding: str,
                     is_training: bool) -> ConvolutionLayer:
        activation = ReLU()
        normalization = BatchNormalization(obj_name="batch_norm",
                                           input_depth=filter_size[-1],
                                           is_training=is_training)
        return cls(obj_name=obj_name,
                   filter_size=filter_size,
                   strides=strides,
                   padding=padding,
                   activation=activation,
                   normalization=normalization,
                   layer_order=["conv", "norm", "activate"])

    def pipeline(self, input_data: tf.Tensor) -> tf.Tensor:
        output = input_data
        for layer_name in self.layer_order:
            output = self.__getattribute__(layer_name)(output)
        return output

    def in_creating(self) -> NoReturn:
        self._filter = ospark.utility.weight_initializer.glorot_uniform(obj_name="filter", shape=self.filter_size)
        self._norm   = self.normalization
        # self.assign(component=ospark.weight.glorot_uniform(obj_name="filter", weight_shape=self.filter_size), name="filter")
        # self.assign(component=self.normalization, name="norm")

    def conv(self, input_data: tf.Tensor) -> tf.Tensor:
        output = tf.nn.conv2d(input=input_data, filters=self._filter, strides=self.strides, padding=self.padding)
        return output

    def norm(self, input_data: tf.Tensor) -> tf.Tensor:
        output = self._norm(input_data)
        return output

    def activate(self, input_data: tf.Tensor) -> tf.Tensor:
        output = self.activation(input_data)
        return output



from . import Metrics
from typing import NoReturn
import numpy as np
import tensorflow as tf
from collections import defaultdict


class ConfusionMatrix(Metrics):

    def __init__(self, class_category: dict) -> NoReturn:
        self._class_category = class_category
        self._counter        = defaultdict(int)

    @property
    def counter(self):
        return self._counter

    @property
    def class_category(self) -> dict:
        return self._class_category

    def process(self, prediction: tf.Tensor, target: tf.Tensor) -> NoReturn:
        prediction_index = tf.argmax(prediction, axis=-1).numpy()
        target_index     = tf.argmax(target, axis=-1).numpy()
        for combined_index in zip(target_index[0], prediction_index[0]):
            self.counter[combined_index] += 1

    def calculate_start(self) -> dict:
        metrics = {}
        category = sorted(self.class_category.keys(), key=lambda x: self.class_category[x])
        basic_matrix = np.zeros(shape=[len(self.class_category), len(self.class_category)])
        for index, count in self.counter.items():
            basic_matrix[index] = count
        print(basic_matrix)
        metrics["accuracy"]  = np.sum(np.diagonal(basic_matrix)) / np.sum(basic_matrix)
        metrics["recall"]    = dict(zip(category, np.diagonal(basic_matrix) / np.sum(basic_matrix, axis=1)))
        metrics["precision"] = dict(zip(category, np.diagonal(basic_matrix) / np.sum(basic_matrix, axis=0)))
        return metrics



from ospark import Model
from typing import List, Optional, Tuple
from ospark.nn.block.transformer_block import transformer_encoder_block
from ospark.nn.layers.dense_layer import DenseLayer
from ospark import weight_initializer
from functools import reduce
import tensorflow as tf
import numpy as np


class VisionTransformer(Model):

    def __init__(self,
                 obj_name: str,
                 image_height: int,
                 image_width: int,
                 patch_height: int,
                 patch_width: int,
                 head_number: int,
                 encoder_number: int,
                 scale_rate: int,
                 dropout_rate: float,
                 encoder_block_number: Optional[int]=None,
                 classification_number: Optional[int]=None,
                 embedding_size: Optional[int]=None,
                 is_training: Optional[bool]=None,
                 trained_weights: Optional[dict]=None):
        super().__init__(obj_name=obj_name, is_training=is_training, trained_weights=trained_weights)
        self._image_height     = image_height
        self._image_width      = image_width
        self._patch_height     = patch_height
        self._patch_width      = patch_width
        assert image_height % patch_height == 0 and image_width % patch_width == 0

        self._head_number      = head_number
        self._scale_rate       = scale_rate
        self._input_dimension  = patch_height * patch_width * 3
        self._embedding_size   = embedding_size or self._input_dimension
        self._dropout_rate     = dropout_rate
        self._positional_table = self.create_positional_encoding_table(max_token=int((image_height / patch_height) * (image_width / patch_width)))

        self._classification_number   = classification_number
        self._linear_projection_layer = DenseLayer(obj_name="linear_projection",
                                                   input_dimension=self._input_dimension,
                                                   hidden_dimension=[self._embedding_size],
                                                   is_training=is_training)

        self._output_layer = DenseLayer(obj_name="output_layer",
                                        input_dimension=self._embedding_size,
                                        hidden_dimension=[self._embedding_size * scale_rate, self._embedding_size],
                                        is_training=is_training)

        if classification_number is not None:
            self._classify_layer = DenseLayer(obj_name="classify_layer",
                                              input_dimension=self._embedding_size,
                                              hidden_dimension=[classification_number],
                                              is_training=is_training)

        self._cls = weight_initializer.glorot_uniform(obj_name="cls_token",
                                                      shape=[1, 1, self._embedding_size],
                                                      trainable=is_training)

        self._blocks = []
        for i in range(encoder_number):
            name = f"block_{i}"
            encoder_block = transformer_encoder_block(obj_name=name,
                                                      embedding_size=self._embedding_size,
                                                      head_number=self._head_number,
                                                      scale_rate=self._scale_rate,
                                                      dropout_rate=self._dropout_rate,
                                                      is_training=is_training)
            self._blocks.append(encoder_block)

    def pipeline(self, input_data: tf.Tensor) -> Tuple[tf.Tensor, tf.Tensor]:
        shape = tf.shape(input_data)

        batch_size, length = shape[0], shape[1]

        input_sequence  = self.process_image(input_data=input_data, batch_size=batch_size)
        input_sequence  = self._linear_projection_layer.pipeline(input_data=input_sequence) # [B, L, D]
        input_sequence += self._positional_table[:, :length, :]
        added_cls       = tf.concat([tf.tile(self._cls, [batch_size, 1, 1]), input_sequence], axis=1)

        encoder_output     = reduce(lambda input_data, block: block.pipeline(input_data=input_data),
                                    self._blocks,
                                    added_cls)

        cls_output    = encoder_output[:, :1, :]
        result        = self._output_layer.pipeline(input_data=cls_output)

        if self._classification_number is not None:
            result = self._classify_layer.pipeline(input_data=result)
        return result

    def process_image(self, input_data: tf.Tensor, batch_size: tf.Tensor) -> tf.Tensor:
        # input_data: [B, H, W, C] -> [B, PN, PN, P, P, C] -> [B, L, D]
        # L = PN * PN, D = P * P * C
        chunking_width  = tf.split(input_data, int(self._image_height / self._patch_height), axis=1)
        chunking_height = reduce(lambda init, chunk: init + tf.split(chunk,
                                                                     int(self._image_width / self._patch_width),
                                                                     axis=2),
                                 chunking_width,
                                 [])

        sequence = self.flatten(input_data=chunking_height, batch_size=batch_size)
        return sequence

    def flatten(self, input_data: List[tf.Tensor], batch_size: tf.Tensor) -> tf.Tensor:
        result = tf.concat([tf.reshape(patch, [batch_size, 1, -1]) for patch in input_data], axis=1)
        return result

    def create_positional_encoding_table(self, max_token: int) -> tf.Tensor:
        basic_table = np.zeros(shape=[max_token, self._embedding_size])
        position    = np.arange(max_token).reshape([-1, 1])
        denominator = np.power(10000, np.arange(0, self._embedding_size, 2) / self._embedding_size)
        basic_table[:, 0::2] = np.sin(position / denominator)
        basic_table[:, 1::2] = np.cos(position / denominator)
        return tf.convert_to_tensor(basic_table, dtype=tf.float32)[tf.newaxis, :, :]



import tensorflow as tf

from ospark.nn.loss_function.loss_function import LossFunction


class Dice(LossFunction):

    def calculate(self, prediction: tf.Tensor, target_data: tf.Tensor) -> tf.Tensor:
        eps = 1e-5
        intersection = tf.reduce_sum(tf.multiply(prediction, target_data))
        union = tf.reduce_sum(prediction) + tf.reduce_sum(target_data) + eps
        loss  = 1. - 2 * intersection / union
        return loss


from ospark.nn.model import Model
from typing import NoReturn, Optional


class Predictor:

    def __init__(self, model: Model):
        self._model = model

    @property
    def model(self) -> Model:
        return self._model

    def predict(self, input_data):
        return NotImplementedError()

    def restore_weights(self, weights):
        return NotImplementedError()

    def __call__(self, input_data):
        prediction = self.predict(input_data=input_data)
        return prediction


from enum import Enum
from typing import Optional
import numpy as np


class PaddingMethod(Enum):

    align_image_center = lambda image, pad_height, pad_width:\
                        np.pad(image, np.array([[np.ceil(pad_height / 2), np.floor(pad_height / 2)],
                                                [np.ceil(pad_width / 2), np.floor(pad_width / 2)],
                                                [0, 0]]).astype(np.int32))
    align_upper_left   = lambda image, pad_height, pad_width:\
                        np.pad(image, np.array([[0, pad_height],
                                                [0, pad_width],
                                                [0, 0]]).astype(np.int32))
    align_upper_right  = lambda image, pad_height, pad_width:\
                        np.pad(image, np.array([[0, pad_height],
                                                [pad_width, 0],
                                                [0, 0]]).astype(np.int32))
    align_lower_left   = lambda image, pad_height, pad_width:\
                        np.pad(image, np.array([[pad_height, 0],
                                                [0, pad_width],
                                                [0, 0]]).astype(np.int32))
    align_lower_right  = lambda image, pad_height, pad_width:\
                        np.pad(image, np.array([[pad_height, 0],
                                                [pad_width, 0],
                                                [0, 0]]).astype(np.int32))


class PaddingManager:

    def __init__(self, padding_method: Optional[str]=None):
        """
        Padding method

        Args:
             padding_method: Optional[str]
                Method has center, upper_left, upper_right, lower_left, lower_right. Default use "center".
        """
        if padding_method == "center" or padding_method is None:
            self._padding = lambda height, width: np.array([[np.ceil(height / 2), np.floor(height / 2)],
                                                            [np.ceil(width / 2), np.floor(width / 2)],
                                                            [0, 0]]).astype(np.int32)
        else:
            y_align, x_align = padding_method.split("_")
            if y_align == "upper":
                y_axis_pad = lambda height: [0, height]
            elif y_align == "lower":
                y_axis_pad = lambda height: [height, 0]
            else:
                raise NameError(f"Check if input {y_align} is wrong")

            if x_align == "left":
                x_axis_pad = lambda width: [0, width]
            elif x_align == "right":
                x_axis_pad = lambda width: [width, 0]
            else:
                raise NameError(f"Check if input {x_align} is wrong")

            self._padding = lambda padding_scale: np.array([y_axis_pad(padding_scale[1]),
                                                            x_axis_pad(padding_scale[0]),
                                                            [0, 0]]).astype(np.int32)

    @property
    def padding(self) -> str:
        return self._padding

    def __call__(self, image: np.ndarray, padding_scale: np.ndarray) -> np.ndarray:
        return np.pad(array=image, pad_width=self.padding(padding_scale))


from ospark.trainer import *
from ospark.nn.loss_function.loss_function import LossFunction
from ospark import Model
import tensorflow as tf
import time


class ExdeepTransformerTrainer(Trainer):

    def __init__(self,
                 data_generator: DataGenerator,
                 model: Model,
                 epoch_number: int,
                 optimizer: Optimizer,
                 loss_function: LossFunction,
                 save_delegate: Optional[SaveDelegate]=None,
                 save_times: Optional[int]=None,
                 save_path: Optional[str]=None,
                 use_profiling_phase: Optional[bool]=True,
                 use_auto_graph: Optional[bool]=True,
                 save_init_weights: Optional[bool]=False,
                 init_weights_path: Optional[str]=None
                 ):
        super().__init__(model=model,
                         data_generator=data_generator,
                         epoch_number=epoch_number,
                         optimizer=optimizer,
                         save_delegate=save_delegate,
                         save_path=save_path,
                         save_times=save_times,
                         use_auto_graph=use_auto_graph,
                         loss_function=loss_function)
        self._use_profiling_phase         = use_profiling_phase
        self._save_init_weights           = save_init_weights
        self._init_weights_path           = init_weights_path or self.save_path.split(".")[0] + "_init.json"
        self._log_file                    = open(self.save_path.split(".")[0] + ".txt", 'w')

    @property
    def use_profiling_phase(self) -> bool:
        return self._use_profiling_phase

    @property
    def log_file(self):
        return self._log_file

    @property
    def save_init_weights(self) -> bool:
        return self._save_init_weights

    @property
    def init_weights_path(self) -> str:
        return self._init_weights_path

    @tf.function(input_signature=[
    tf.TensorSpec(shape=(None, None), dtype=tf.int64),
    tf.TensorSpec(shape=(None, None), dtype=tf.int64)])
    def graph_mode(self,
                   train_data: tf.Tensor,
                   target_data: tf.Tensor):
        return self.train_step(train_data=train_data, target_data=target_data)

    def start(self) -> NoReturn:
        if self.save_init_weights:
            weights = self.weights_operator.weights
            self.save(weights=weights, path=self.init_weights_path)

        if self.use_profiling_phase:
            dataset = next(iter(self.data_generator))
            profiling_encoder_input, profiling_decoder_input = dataset.training_data, dataset.target_data

            print("Profiling phase start.")
            self.model.profiling_phase(encoder_input=profiling_encoder_input,
                                       decoder_input=profiling_decoder_input[:, :-1])
            print("Profiling phase end.")

            print("=" * 24)
        else:
            self.model.back_to_standard()
        print("Training phase start.")
        self.training_process(epoch_number=self.epoch_number)
        print("Training phase end.")
        print("=" * 24)

    def training_process(self, epoch_number) -> NoReturn:
        for epoch in range(epoch_number):
            total_loss_value = 0
            training_count   = 0
            total_accuracies = 0
            start_time       = time.time()
            for batch, dataset in enumerate(self.data_generator):
                training_data, target_data = dataset.training_data, dataset.target_data
                accuracies, loss_value = self.training_method(training_data, target_data)
                total_accuracies += accuracies
                total_loss_value += loss_value
                training_count   += 1
                print(loss_value)
                if (batch + 1) % 50 == 0:
                    print(f"Epoch {epoch + 1} Batch {batch} Loss {loss_value:.4f} Accuracy {accuracies:.4f}")

            print(f'Epoch {epoch + 1} '
                  f'Loss {total_loss_value / training_count:.4f} '
                  f'Accuracy {total_accuracies / training_count:.4f}', file=self.log_file)
            print(f'Epoch {epoch + 1} '
                  f'Loss {total_loss_value / training_count:.4f} '
                  f'Accuracy {total_accuracies / training_count:.4f}')
            print(f'Time taken for 1 epoch: {time.time() - start_time:.2f} secs\n')
            if self.will_save(epoch_number=epoch):
                self.save_delegate.save(weights=self.weights_operator.weights, path=self.save_path)
        self.save(weights=self.weights_operator.weights, path=self.save_path)
        self.log_file.close()

    def calculate_accuracy(self, prediction: tf.Tensor, target: tf.Tensor) -> tf.Tensor:
        accuracies = tf.math.equal(target, tf.argmax(prediction, axis=-1))
        mask       = tf.logical_not(tf.math.equal(target, 0))
        accuracies = tf.logical_and(mask, accuracies)
        accuracies = tf.cast(accuracies, tf.float32)
        mask       = tf.cast(mask, tf.float32)
        return tf.reduce_sum(accuracies) / tf.reduce_sum(mask)

    def train_step(self, train_data: tf.Tensor, target_data: tf.Tensor):
        with tf.GradientTape() as tape:
            weights    = self.model.training_weights
            prediction = self.model.pipeline(encoder_input=train_data, decoder_input=target_data[:, :-1])
            loss_value = self.loss_function(prediction=prediction, target_data=target_data[:, 1:])
            accuracies = self.calculate_accuracy(prediction=prediction, target=target_data[:, 1:])
            tape.watch(weights)
        gradients  = tape.gradient(loss_value, weights)
        self.optimizer.apply_gradients(zip(gradients, weights))
        return accuracies, loss_value




from ospark.models.informer import Informer
from ospark.models.transformer import Transformer
from ospark.models.exdeep_transformer import ExdeepTransformer
from ospark.nn.block.informer_block import informer_decoder_block, informer_encoder_block
from ospark.nn.block.transformer_block import transformer_encoder_block, transformer_decoder_block, performer_decoder_block, performer_encoder_block
from ospark.nn.block.deep_transformer_block import exdeep_decoder_block, exdeep_encoder_block
from typing import Optional


class FormerBuilder:

    @staticmethod
    def informer(class_number: int,
                 block_number: int,
                 embedding_size: int,
                 head_number: int,
                 scale_rate: int,
                 sample_factor: float,
                 dropout_rate: float,
                 trained_weights: Optional[dict] = None,
                 is_training: Optional[bool] = None,
                 use_decoder: bool = True,
                 filter_width: int = None,
                 pooling_size: list = None,
                 strides: list = None
                 ) -> Informer:
            encoder_blocks = []
            decoder_blocks = []
            for i in range(block_number):
                encoder_name = f"encoder_block_{i}"
                encoder_blocks.append(informer_encoder_block(obj_name=encoder_name,
                                                             embedding_size=embedding_size,
                                                             head_number=head_number,
                                                             scale_rate=scale_rate,
                                                             sample_factor=sample_factor,
                                                             filter_width=filter_width,
                                                             pooling_size=pooling_size,
                                                             strides=strides,
                                                             is_training=is_training))
                if use_decoder:
                    decoder_name = f"decoder_block_{i}"
                    decoder_blocks.append(informer_decoder_block(obj_name=decoder_name,
                                                                 embedding_size=embedding_size,
                                                                 head_number=head_number,
                                                                 scale_rate=scale_rate,
                                                                 sample_factor=sample_factor,
                                                                 is_training=is_training))
            return Informer(obj_name="Informer",
                            trained_weights=trained_weights,
                            encoder_blocks=encoder_blocks,
                            class_number=class_number,
                            embedding_size=embedding_size,
                            decoder_blocks=decoder_blocks,
                            dropout_rate=dropout_rate,
                            is_training=is_training)

    @staticmethod
    def transformer(block_number: int,
                    head_number: int,
                    embedding_size: int,
                    scale_rate: int,
                    class_number: int,
                    dropout_rate: float,
                    trained_weights: Optional[dict] = None,
                    is_training: Optional[bool] = None,
                    encoder_corpus_size: Optional[int] = None,
                    decoder_corpus_size: Optional[int] = None,
                    use_embedding_layer: Optional[bool] = True,
                    max_length: Optional[int] = 2000,
                    ) -> Transformer:
        encoder_blocks = []
        decoder_blocks = []
        for i in range(block_number):
            encoder_name = f"encoder_block_{i}"
            encoder_blocks.append(transformer_encoder_block(obj_name=encoder_name,
                                                            embedding_size=embedding_size,
                                                            head_number=head_number,
                                                            scale_rate=scale_rate,
                                                            dropout_rate=dropout_rate,
                                                            is_training=is_training))
            decoder_name = f"decoder_block_{i}"
            decoder_blocks.append((transformer_decoder_block(obj_name=decoder_name,
                                                             embedding_size=embedding_size,
                                                             head_number=head_number,
                                                             scale_rate=scale_rate,
                                                             dropout_rate=dropout_rate,
                                                             is_training=is_training)))
        return Transformer(obj_name="Transformer",
                           trained_weights=trained_weights,
                           encoder_blocks=encoder_blocks,
                           class_number=class_number,
                           decoder_blocks=decoder_blocks,
                           max_length=max_length,
                           encoder_corpus_size=encoder_corpus_size,
                           decoder_corpus_size=decoder_corpus_size,
                           use_embedding_layer=use_embedding_layer,
                           embedding_size=embedding_size,
                           dropout_rate=dropout_rate,
                           is_training=is_training)

    @staticmethod
    def exdeep_transformer(encoder_block_number: int,
                           decoder_block_number: int,
                           head_number: int,
                           embedding_size: int,
                           scale_rate: int,
                           class_number: int,
                           dropout_rate: float,
                           trained_weights: Optional[dict] = None,
                           is_training: Optional[bool] = None,
                           encoder_corpus_size: Optional[int] = None,
                           decoder_corpus_size: Optional[int] = None,
                           use_embedding_layer: Optional[bool] = True,
                           max_length: Optional[int] = 2000,
                           ) -> Transformer:
        encoder_blocks = []
        decoder_blocks = []
        for i in range(encoder_block_number):
            encoder_name = f"encoder_block_{i}"
            encoder_blocks.append(exdeep_encoder_block(obj_name=encoder_name,
                                                       embedding_size=embedding_size,
                                                       head_number=head_number,
                                                       scale_rate=scale_rate,
                                                       dropout_rate=dropout_rate,
                                                       is_training=is_training,))
        for i in range(decoder_block_number):
            decoder_name = f"decoder_block_{i}"
            decoder_blocks.append((exdeep_decoder_block(obj_name=decoder_name,
                                                        embedding_size=embedding_size,
                                                        head_number=head_number,
                                                        scale_rate=scale_rate,
                                                        dropout_rate=dropout_rate,
                                                        is_training=is_training)))
        return ExdeepTransformer(obj_name="ExdeepTransformer",
                                 trained_weights=trained_weights,
                                 encoder_blocks=encoder_blocks,
                                 class_number=class_number,
                                 decoder_blocks=decoder_blocks,
                                 max_length=max_length,
                                 encoder_corpus_size=encoder_corpus_size,
                                 decoder_corpus_size=decoder_corpus_size,
                                 use_embedding_layer=use_embedding_layer,
                                 embedding_size=embedding_size,
                                 dropout_rate=dropout_rate,
                                 is_training=is_training)

    @staticmethod
    def performer(block_number: int,
                  head_number: int,
                  embedding_size: int,
                  scale_rate: int,
                  class_number: int,
                  dropout_rate: float,
                  random_projections_number: int,
                  trained_weights: Optional[dict] = None,
                  is_training: Optional[bool] = False,
                  encoder_corpus_size: Optional[int] = None,
                  decoder_corpus_size: Optional[int] = None,
                  use_embedding_layer: Optional[bool] = True,
                  max_length: Optional[int] = 2000,
                  ) -> Transformer:
        encoder_blocks = []
        decoder_blocks = []
        for i in range(block_number):
            encoder_name = f"encoder_block_{i}"
            encoder_blocks.append(performer_encoder_block(obj_name=encoder_name,
                                                          embedding_size=embedding_size,
                                                          head_number=head_number,
                                                          scale_rate=scale_rate,
                                                          dropout_rate=dropout_rate,
                                                          is_training=is_training,
                                                          random_projections_number=random_projections_number))
            decoder_name = f"decoder_block_{i}"
            decoder_blocks.append((performer_decoder_block(obj_name=decoder_name,
                                                           embedding_size=embedding_size,
                                                           head_number=head_number,
                                                           scale_rate=scale_rate,
                                                           dropout_rate=dropout_rate,
                                                           is_training=is_training,
                                                           random_projections_number=random_projections_number)))
        return Transformer(obj_name="Transformer",
                           trained_weights=trained_weights,
                           encoder_blocks=encoder_blocks,
                           class_number=class_number,
                           decoder_blocks=decoder_blocks,
                           max_length=max_length,
                           encoder_corpus_size=encoder_corpus_size,
                           decoder_corpus_size=decoder_corpus_size,
                           use_embedding_layer=use_embedding_layer,
                           embedding_size=embedding_size,
                           dropout_rate=dropout_rate,
                           is_training=is_training)



from __future__ import annotations
from ospark.data.generator import DataGenerator
from ospark.utility.padding_method import PaddingManager
from typing import List, Optional, Set, Tuple
from ospark.data.path import DataPath
from ospark.data.folder import DataFolder
from PIL import Image
import tensorflow as tf
import numpy as np
import pathlib
import random
import copy
import cv2
import os


class FOTSDataGenerator(DataGenerator):

    class Dataset:

        def __init__(self):
            self._training_data = None
            self._target_data   = None
            self._words         = None
            self._bbox_points   = None

        @property
        def training_data(self) -> tf.Tensor:
            return self._training_data

        @property
        def target_data(self) -> tf.Tensor:
            return self._target_data

        @property
        def words(self) -> List[List[str]]:
            return self._words

        @property
        def bbox_points(self) -> List[List[np.ndarray]]:
            return self._bbox_points

        def data_setting(self,
                         training_data: tf.Tensor,
                         target_data: tf.Tensor,
                         words: List[List[str]],
                         bbox_points: List[List[np.ndarray]]):
            self._training_data = training_data
            self._target_data   = target_data
            self._words         = words
            self._bbox_points   = bbox_points

    def __init__(self,
                 data_folder: DataFolder,
                 batch_size: int,
                 height_threshold: int,
                 ignore_words: Set[str],
                 target_size: list,
                 image_shrink_scale: float,
                 padding_method: Optional[str]="center",
                 bbox_shrink_scale: Optional[float]=0.3,
                 initial_step: Optional[int]=None,
                 use_shuffle: Optional[bool]=False):
        """

        :param training_folder:
        :param labeling_folder:
        :param batch_size:
        :param height_threshold:
        :param ignore_words:
        image_size: np.ndarray
            image size [width, height]
        :param image_shrink_scale:
        padding_method: str
            Method has center, upper_left, upper_right, lower_left, lower_right. Default use "center".
        :param bbox_shrink_scale:
        :param initial_step:
        """

        self._folder                = data_folder
        self._height_threshold      = height_threshold
        self._ignore_words          = ignore_words
        self._shrink_image          = image_shrink_scale
        self._bbox_shrink_scale     = bbox_shrink_scale
        self._target_size           = np.array(target_size)
        self._padding               = PaddingManager(padding_method)
        self._image_shrink_size     = (np.array(target_size).astype(np.float32) * image_shrink_scale).astype(np.int32)
        self._blank_image           = np.zeros(shape=[self.image_shrink_size[1], self.image_shrink_size[0], 1])
        self._coordinate_matrix     = self.create_coordinate_matrix()
        self._basic_move_distance   = np.array([[1, 1], [-1, 1], [-1, -1], [1, -1]])
        self._data_indices          = [i for i in self._folder.indexed_training_data.keys()]
        self._data_path             = DataPath(batch_size=batch_size,
                                               total_data=len(self._folder.training_files),
                                               index_table=list(self.folder.indexed_training_data.keys()))
        self._shuffle               = use_shuffle
        if self.shuffle:
            random.shuffle(self._data_indices)

        super().__init__(training_data=self._folder.training_files,
                         target_data=self._folder.labeling_files,
                         batch_size=batch_size,
                         initial_step=initial_step)

    @property
    def folder(self) -> DataFolder:
        return self._folder

    @property
    def height_threshold(self) -> int:
        return self._height_threshold

    @property
    def ignore_words(self) -> Set[str]:
        return self._ignore_words

    @property
    def shrink_image(self) -> float:
        return self._shrink_image

    @property
    def bbox_shrink_scale(self) -> float:
        return self._bbox_shrink_scale

    @property
    def target_size(self) -> np.ndarray:
        return self._target_size

    @property
    def padding(self) -> PaddingManager:
        return self._padding

    @property
    def image_shrink_size(self) -> np.ndarray:
        return self._image_shrink_size

    @property
    def blank_image(self) -> np.ndarray:
        return self._blank_image

    @property
    def coordinate_matrix(self) -> np.ndarray:
        return self._coordinate_matrix

    @property
    def data_indices(self) -> list:
        return self._data_indices

    @property
    def shuffle(self) -> bool:
        return self._shuffle

    def create_coordinate_matrix(self) -> np.ndarray:
        """
        This function doing create coordinate matrix.

        Returns:
             coordinate_matrix: np.ndarray
                shape is [y_size, x_size, axis_number]
        """
        x_axis = np.tile(np.arange(self.image_shrink_size[0])[np.newaxis, :, np.newaxis], [self.image_shrink_size[1], 1, 1])
        y_axis = np.tile(np.arange(self.image_shrink_size[1])[:, np.newaxis, np.newaxis], [1, self.image_shrink_size[0], 1])
        return np.concatenate([x_axis, y_axis], axis=-1)

    def __iter__(self) -> FOTSDataGenerator:
        return self

    def __next__(self) -> Dataset:
        if self.step < self.max_step:
            dataset = self._get_data()
            self._step += 1
            return dataset
        self.reset()
        raise StopIteration()

    def _get_data(self) -> Dataset:
        for index in self._data_path.range(step=self.step):
            training_data_path, labeling_data_path = self.folder.get_files(index=index)
            self._data_path.training_data_paths = training_data_path
            self._data_path.label_data_paths    = labeling_data_path
        training_data = self.process_training_data(self._data_path.training_data_paths)
        target_image, words, bbox_points = self.process_target_data(self._data_path.label_data_paths)
        self.dataset.data_setting(training_data=training_data,
                                  target_data=target_image,
                                  words=words,
                                  bbox_points=bbox_points)
        return self.dataset

    def process_training_data(self, paths: List[str]) -> tf.Tensor:
        imgs = []
        for path in paths:
            img = Image.open(path)
            img = self.padding_and_resize(img)
            imgs.append(img[np.newaxis, :, :, :])
        return tf.cast(tf.concat(imgs, axis=0), dtype=tf.float32)

    def padding_and_resize(self, img: Image.Image) -> np.ndarray:
        input_size = np.array(img.size)
        if all(input_size != self.target_size):
            resize_scale  = (input_size / max(input_size / self.target_size)).astype(np.int32)
            padding_scale = self.target_size - resize_scale
            img = img.resize(size=resize_scale, resample=Image.BILINEAR)
            img = self.padding(image=np.array(img), padding_scale=padding_scale)
        else:
            img = np.array(img)
        return img

    def process_target_data(self, paths: List[str]) -> Tuple[tf.Tensor, List[List[str]], List[List[np.ndarray]]]:
        target_images       = []
        words_stacked       = []
        bbox_points_stacked = []
        for path in paths:
            detection_map = np.zeros(shape=[self.image_shrink_size[1], self.image_shrink_size[0], 6])
            fp = open(path, 'r', encoding="utf-8-sig")
            points = []
            words  = []
            for datasets in fp:
                target      = datasets.strip().split(",")
                bbox_points = np.reshape(np.around(np.array([int(j) for j in target[:8]]) * self.shrink_image).astype(np.int32),
                                         [-1, 2])
                _height      = np.linalg.norm(bbox_points[-1] - bbox_points[0], axis=0)
                _width       = np.linalg.norm(bbox_points[0] - bbox_points[1], axis=0)
                if _height > 2 * _width:
                    height      = _width
                    width       = _height
                    bbox_points = bbox_points[[1, 2, 3, 0], :]
                else:
                    height      = _height
                    width       = _width

                # filtered according origin image size.
                if height / self.shrink_image < self.height_threshold or width / self.shrink_image < self.height_threshold:
                    continue

                detection_map += self.create_detection_map(bbox_points=bbox_points)
                target_word    = ",".join(target[8:])

                if target_word in self.ignore_words:
                    continue
                points.append(bbox_points)
                words.append(target_word)
            target_images.append(detection_map[np.newaxis, :, :, :])
            if points != []:
                words_stacked.append(words)
                bbox_points_stacked.append(points)
        return tf.cast(tf.concat(target_images, axis=0), dtype=tf.float32), words_stacked, bbox_points_stacked

    def create_detection_map(self, bbox_points: np.ndarray) -> np.ndarray:
        shrunk_points   = self.create_shrunk_coordinate(bbox_points)
        score_map       = cv2.fillPoly(copy.copy(self.blank_image), [shrunk_points], 1)
        bbox_map, angle = self.create_bbox_map(score_map=score_map, bbox_points=bbox_points)
        angle_map       = angle * score_map
        return np.concatenate([score_map, bbox_map, angle_map], axis=-1)

    def create_shrunk_coordinate(self, bbox_points: np.ndarray) -> np.ndarray:
        clockwise_vector        = (bbox_points[[1, 2, 3, 0], :] - bbox_points) * self.bbox_shrink_scale / 2
        counterclockwise_vector = (bbox_points[[3, 0, 1, 2], :] - bbox_points) * self.bbox_shrink_scale / 2
        shrunk_vector           = clockwise_vector + counterclockwise_vector
        shrunk_points           = np.round(bbox_points + shrunk_vector).astype(np.int32)
        equal_position          = np.where(shrunk_points == bbox_points)

        shrunk_points[equal_position] += self._basic_move_distance[equal_position]
        return shrunk_points

    def create_bbox_map(self, score_map: np.ndarray, bbox_points: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        score_map_coordinate = score_map * self.coordinate_matrix
        a, b, c, angle       = self.calculate_coefficient(bbox_points)
        distance = self.calculate_distance(coordinate_matrix=score_map_coordinate, a=a, b=b, c=c)
        # 因為平行四邊形的關係，所以除以 np.cos(angle)
        distance[[1, 3], :, :] = distance[[1, 3], :, :] / np.cos(angle)
        distance = np.transpose(distance, [1, 2, 0]) * score_map
        return distance, angle

    def calculate_coefficient(self, bbox_points: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """
        linear equation: ax + by + c = 0
        equation coefficient is a, b and c.
        """

        x     = bbox_points[[1, 2, 3, 0], 0] - bbox_points[[0, 1, 2, 3], 0] + 1e-9
        y     = bbox_points[[1, 2, 3, 0], 1] - bbox_points[[0, 1, 2, 3], 1]
        a     = y / x
        b     = np.array(1)
        c     = bbox_points[[1, 2, 3, 0], 1] - a * bbox_points[[1, 2, 3, 0], 0]
        angle = self.calculate_angle(bbox_points[0], bbox_points[1])
        return -1 * a[:, np.newaxis, np.newaxis], b, -1 * c[:, np.newaxis, np.newaxis], angle

    def calculate_distance(self,
                           coordinate_matrix: np.ndarray,
                           a: np.ndarray,
                           b: np.ndarray,
                           c: np.ndarray) -> np.ndarray:
        """
        linear equation: ax + by + c = 0
        equation coefficient is a, b and c.
        """

        denominator = np.sqrt(np.power(a, 2) + 1)
        numerator   = np.abs(a * coordinate_matrix[:, :, 0] + b * coordinate_matrix[:, :, 1] + c)
        return numerator / denominator

    def calculate_angle(self,
                        start_point: np.ndarray,
                        end_point: np.ndarray) -> np.ndarray:
        vector = end_point - start_point
        length = np.linalg.norm(vector)

        x_partition, y_partition = vector

        coefficient = 1 if y_partition >= 0 else -1
        angle = np.arccos(x_partition / length) / coefficient
        return angle


from ospark.nn.layers.activation import PassActivation, ReLU, Swish, Sigmoid, Activation
from ospark import weight_initializer
from typing import Optional, NoReturn
import tensorflow as tf


class Bilinear(Activation):

    def __init__(self,
                 dimension: int,
                 is_training: Optional[bool]):
        super(Bilinear, self).__init__(is_training=is_training)
        self._weight_1 = weight_initializer.glorot_uniform(obj_name="weight_1", shape=[dimension, dimension])
        self._weight_2 = weight_initializer.glorot_uniform(obj_name="weight_2", shape=[dimension, dimension])

        self._bias_1 = weight_initializer.zeros(obj_name="bias_1", shape=[dimension])
        self._bias_2 = weight_initializer.zeros(obj_name="bias_2", shape=[dimension])

        self.activation_function = PassActivation()

    @property
    def activation_function(self) -> Activation:
        return self._activation_function

    @activation_function.setter
    def activation_function(self, value: Activation) -> NoReturn:
        self._activation_function = value

    def pipeline(self, input_data: tf.Tensor) -> tf.Tensor:
        linear_layer_1 = self.activation_function(tf.matmul(input_data, self._weight_1) + self._bias_1)
        linear_layer_2 = tf.matmul(input_data, self._weight_2) + self._bias_2
        return tf.multiply(linear_layer_1, linear_layer_2)


class GLU(Bilinear):

    def __init__(self,
                 dimension: int,
                 is_training: Optional[bool]):
        super(GLU, self).__init__(dimension=dimension, is_training=is_training)

        self.activation_function = Sigmoid()

class ReGLU(GLU):

    def __init__(self,
                 dimension: int,
                 is_training: Optional[bool]):
        super(ReGLU, self).__init__(dimension=dimension, is_training=is_training)

        self.activation_function = ReLU()

class SwiGLU(GLU):

    def __init__(self,
                 dimension: int,
                 beta: Optional[float]=None,
                 is_training: Optional[bool]=None):
        super(SwiGLU, self).__init__(dimension=dimension, is_training=is_training)

        self.activation_function = Swish(beta=beta or 1.)





import ospark.utility.weight_initializer
from . import Layer
from typing import NoReturn, Optional
from ospark.nn.layers.activation import Activation
from ospark.nn.layers.normalization import Normalization, BatchNormalization
import tensorflow as tf
import ospark


class DistillingLayer(Layer):

    def __init__(self, 
                 obj_name: str,
                 embedding_size: int,
                 is_training: Optional[bool]=None,
                 filter_width: Optional[int]=None,
                 pooling_size: Optional[list]=None,
                 strides: Optional[list]=None,
                 activation: Optional[Activation]=None,
                 norm: Optional[Normalization]=None):
        super().__init__(obj_name=obj_name, is_training=is_training)
        self._embedding_size = embedding_size
        self._filter_width   = filter_width or 3
        self._pooling_size   = pooling_size or [1, 3, 1]
        self._strides        = strides or [1, 2, 1]
        self._norm           = norm or BatchNormalization(obj_name="batch_norm",
                                                          input_depth=embedding_size,
                                                          is_training=is_training)
        self._activation     = activation or ospark.nn.layers.activation.ELU()
        self._pooling        = tf.nn.max_pool1d

    @property
    def embedding_size(self) -> int:
        return self._embedding_size

    @property
    def activation(self) -> Activation:
        return self._activation

    @property
    def pooling(self):
        return self._pooling

    @property
    def strides(self) -> list:
        return self._strides

    @property
    def filter_width(self) ->int:
        return self._filter_width

    @property
    def pooling_size(self) -> list:
        return self._pooling_size

    @property
    def norm(self) -> Normalization:
        return self._norm

    def in_creating(self) -> NoReturn:
        self._conv_filter = ospark.utility.weight_initializer.truncated_normal(obj_name="conv_filter",
                                                                               shape=[self.filter_width,
                                                                  self.embedding_size,
                                                                  self.embedding_size])
        self._batch_norm  = self.norm
        # self.assign(component=ospark.weight.truncated_normal(obj_name="conv_filter",
        #                                                      weight_shape=[self.filter_width,
        #                                                                    self.embedding_size,
        #                                                                    self.embedding_size]))
        # self.assign(component=self.norm, name="norm")

    def pipeline(self, input_data: tf.Tensor) -> tf.Tensor:
        conv_output = tf.nn.conv1d(input_data, self._conv_filter, [1, 1, 1], padding="SAME")
        norm_output = self._batch_norm(conv_output)
        activation  = self.activation(norm_output)
        pooling     = self.pooling(activation, self.pooling_size, self.strides, padding="SAME")
        return pooling


from __future__ import annotations
from ospark.nn.component.weight import Weight
from typing import NoReturn, Optional, Set, Union, Any
from .name_space import NameSpace
import tensorflow as tf


# TODO 如果這邊 input_data 內有 Weight 的話，會有問題，因為沒辦法用 Descriptor 去 get
def dismantling_nest_structure(input_data: list, obj: ModelObject):
    for i, inside_obj in enumerate(input_data):
        if isinstance(inside_obj, (ModelObject, Weight)):
            obj.assign(inside_obj)
        elif isinstance(inside_obj, list):
            dismantling_nest_structure(input_data=inside_obj, obj=obj)
        else:
            continue


class Descriptor:

    def __init__(self):
        self._obj_package = {}

    def __get__(self, instance: Any, instance_type: Any) -> Union[tf.Tensor, ModelObject]:
        identifier = id(instance)
        if isinstance(self._obj_package[identifier], Weight):
            return self._obj_package[identifier].value
        else:
            return self._obj_package[identifier]

    def __set__(self, instance: ModelObject, value: Any) -> NoReturn:
        if isinstance(value, (ModelObject, Weight)):
            instance.assign(value)
            if isinstance(value, Weight) and value._is_default_setting:
                value._trainable = instance.is_training
            elif isinstance(value, ModelObject):
                value._is_training = instance.is_training
        identifier = id(instance)
        self._obj_package[identifier] = value


class MetaObject(type):

    def __call__(cls, *args, **kwargs) -> ModelObject:
        obj = super().__call__(*args, **kwargs)
        obj.in_creating()

        keys = list(obj.__dict__.keys())
        for attr_name in keys:
            attr_value = obj.__dict__[attr_name]
            if isinstance(attr_value, (ModelObject, Weight)):
                sub_obj = obj.__dict__.pop(attr_name)
                if attr_name not in cls.__dict__:
                    setattr(cls, attr_name, Descriptor())
                setattr(obj, attr_name, sub_obj)
            elif isinstance(attr_value, list):
                dismantling_nest_structure(input_data=attr_value, obj=obj)
        return obj


class ModelObject(metaclass=MetaObject):

    def __init__(self, obj_name: str, is_training: Optional[bool]=None, training_phase: Optional[bool]=None):
        self._obj_name       = obj_name
        self._assigned       = Assigned()
        self._training_phase = training_phase if training_phase is not None else True
        self._is_training    = is_training if is_training is not None else True

    @property
    def obj_name(self) -> str:
        return self._obj_name

    @property
    def assigned(self) -> Assigned:
        return self._assigned

    @property
    def is_training(self) -> bool:
        return self._is_training

    @property
    def training_phase(self) -> bool:
        return self._training_phase

    @training_phase.setter
    def training_phase(self, value: bool) -> None:
        self._training_phase = value
        for assigned in self.assigned:
            if isinstance(assigned, ModelObject):
                assigned.training_phase = value

    def in_creating(self) -> NoReturn:
        pass

    def assign(self, component: Union[ModelObject, Weight], name: Optional[str]=None) -> Union[ModelObject, Weight]:
        self.assigned.assign(component, name)
        return component

    def create(self) -> NoReturn:
        with NameSpace(name=self.obj_name):
            for component in self.assigned:
                component.create()


class Assigned:

    def __init__(self) -> NoReturn:
        self._component_names = set()

    @property
    def component_names(self) -> Set[str]:
        return self._component_names

    def assign(self, component: ModelObject, name: Optional[str]=None) -> NoReturn:
        if name is None:
            name = component.obj_name
        setattr(self, name, component)
        self._component_names.add(name)

    def __getattr__(self, name):
        if name != "is_tensor_like":
            return self.__dict__[name]

    def __getattribute__(self, name: str):
        obj = super().__getattribute__(name)
        if isinstance(obj, Weight):
            return obj.value
        else:
            return obj

    def __iter__(self) -> ModelObject:
        for component_name in self.component_names:
            obj = self.__getattr__(component_name)
            yield obj


from ospark.backbone.backbone import Backbone
from ospark.nn.block import Block
from typing import NoReturn
import tensorflow as tf


class Autoencoder(Backbone):

    def __init__(self,
                 obj_name: str,
                 encoder: Block,
                 decoder: Block):
        super().__init__(obj_name=obj_name)
        self._encoder   = encoder
        self._decoder   = decoder

    @property
    def encoder(self) -> Block:
        return self._encoder

    @property
    def decoder(self) -> Block:
        return self._decoder

    def in_creating(self) -> NoReturn:
        self.assign(component=self.encoder, name="encoder")
        self.assign(component=self.decoder, name="decoder")

    def pipeline(self, input_data: tf.Tensor) -> tf.Tensor:
        encoder_output = self.assigned.encoder.pipeline(input_data)
        decoder_output = self.assigned.decoder.pipeline(encoder_output)
        return decoder_output


import tensorflow as tf

from ospark.nn.loss_function.loss_function import LossFunction


class Balanced(LossFunction):

    def __init__(self, image_width: int, image_height: int, omega: int):
        self._image_width  = image_width
        self._image_height = image_height
        self._omega        = omega

    @property
    def image_width(self) -> int:
        return self._image_width

    @property
    def image_height(self) -> int:
        return self._image_height

    @property
    def omega(self) -> int:
        return self._omega

    def calculate_loss(self, prediction: tf.Tensor, target_data: tf.Tensor) -> tf.Tensor:
        total_elements = self.image_width * self.image_height
        beta = 1 - (self.omega / total_elements)
        reward_part = tf.multiply(beta, tf.multiply(target_data, tf.math.log(prediction)))
        punish_part = tf.multiply((1. - beta), tf.multiply((1. - target_data), tf.math.log((1. - prediction))))
        loss = -tf.reduce_mean(tf.add(reward_part, punish_part))
        return loss


from ospark.trainer.exdeep_transformer import ExdeepTransformerTrainer
from ospark.data.generator.translate_data_generator import TranslateDataGenerator
from ospark.models.builder import FormerBuilder
from ospark.predictor.translator import Translator
from typing import Optional
from ospark.data.encoder import LanguageDataEncoder
from sacrebleu.metrics import BLEU
import tensorflow_datasets as tfds
import numpy as np
import json
import os

def text_encoder(folder_path: str,
                 vocab_file: str,
                 vocabulary_size: int,
                 datasets: Optional[list]=None) -> tfds.deprecated.text.SubwordTextEncoder:
    file = os.path.join(folder_path, vocab_file)
    try:
        subword_encoder = tfds.deprecated.text.SubwordTextEncoder.load_from_file(file)
        print(f"載入已建立的字典： {file}")
    except:
        print("沒有已建立的字典，從頭建立。")
        subword_encoder = tfds.deprecated.text.SubwordTextEncoder.build_from_corpus(
            datasets,
            target_vocab_size=vocabulary_size)
        # 將字典檔案存下以方便下次 warm start
        subword_encoder.save_to_file(file)
    return subword_encoder


weight_path = "/Users/abnertsai/Documents/Ospark/ospark/samples"
file_name   = "exdeep_16_50_60_12_4_128.json"
# file_name     = "exdeep_32_50_6_6_8_512.json"


folder_path          = "/Users/abnertsai/Documents/Gitlab/self-attention/tensorflow_datasets"
# folder_path          = os.path.dirname(os.path.abspath(__file__))
dataset_name         = 'wmt14_translate/de-en'
train_vocab_file     = "de_vocab_file"
target_vocab_file    = "en_vocab_file"
vocabulary_size      = 40000
scale_rate           = 4

model, batch_size, epoch_number, encoder_block_number, decoder_block_number, head_number, embedding_size = file_name.split(".")[0].split("_")

with open(os.path.join(weight_path, file_name), 'r') as fp:
    weights = json.load(fp)


ds = tfds.load(data_dir=folder_path,
               name=dataset_name,
               as_supervised=True)

train_data = ds["train"]



train_datasets, target_datasets = None, None
if not os.path.isfile(os.path.join(folder_path, train_vocab_file + ".subwords")):
    print("找不到 text encoder file，讀取 datasets 建立 text encoder")
    train_datasets, target_datasets = zip(*[[train_data.numpy(), target_data.numpy()]
                                            for train_data, target_data
                                            in train_data])


# 建立/讀取 text_encoder
train_data_text_encoder = text_encoder(folder_path=folder_path,
                                       vocab_file=train_vocab_file,
                                       vocabulary_size=vocabulary_size,
                                       datasets=train_datasets)

target_data_text_encoder = text_encoder(folder_path=folder_path,
                                        vocab_file=target_vocab_file,
                                        vocabulary_size=vocabulary_size,
                                        datasets=target_datasets)

exdeep_model = FormerBuilder.exdeep_transformer(encoder_block_number=int(encoder_block_number),
                                                decoder_block_number=int(decoder_block_number),
                                                head_number=int(head_number),
                                                embedding_size=int(embedding_size),
                                                scale_rate=scale_rate,
                                                class_number=target_data_text_encoder.vocab_size + 2,
                                                encoder_corpus_size=train_data_text_encoder.vocab_size + 2,
                                                decoder_corpus_size=target_data_text_encoder.vocab_size + 2)


data_encoder   = LanguageDataEncoder(train_data_encoder=train_data_text_encoder,
                                     label_data_encoder=target_data_text_encoder)
# 建立 data_generator
data_generator = TranslateDataGenerator(datasets=train_data,
                                        data_encoder=data_encoder,
                                        batch_size=1,
                                        take_number=None,
                                        max_length=None)

predictor = Translator(model=exdeep_model,
                       input_text_encoder=train_data_text_encoder,
                       output_text_encoder=target_data_text_encoder,
                       predict_max_length=100)

predictor.restore_weights(weights=weights)





from typing import List, Optional
import tensorflow as tf
from ospark import Weight


class Glorot(Weight):

    def init_weight(self) -> tf.Tensor:
        limit  = tf.cast(tf.sqrt(6 / (self.shape[-1] + self.shape[-2])), dtype=tf.float32)
        weight = tf.random.uniform(self.shape, minval=-limit, maxval=limit)
        return weight


class Zeros(Weight):

    def init_weight(self) -> tf.Tensor:
        weight = tf.zeros(self.shape)
        return weight


class Ones(Weight):

    def init_weight(self) -> tf.Tensor:
        weight = tf.ones(shape=self.shape)
        return weight


class TruncatedNormal(Weight):

    def init_weight(self) -> tf.Tensor:
        weight = tf.random.truncated_normal(shape=self.shape)
        return weight


class Normal(Weight):

    def init_weight(self) -> tf.Tensor:
        weight = tf.random.normal(shape=self.shape)
        return weight


class Uniform(Weight):

    def init_weight(self) -> tf.Tensor:
        weight = tf.random.uniform(shape=self.shape)
        return weight


def truncated_normal(obj_name: str, shape: List[int], trainable: Optional[bool]=None) -> Weight:
    return TruncatedNormal(obj_name=obj_name, shape=shape, trainable=trainable)


def normal(obj_name: str, shape: List[int], trainable: Optional[bool]=None) -> Weight:
    return Normal(obj_name=obj_name, shape=shape, trainable=trainable)


def uniform(obj_name: str, shape: List[int], trainable: Optional[bool]=None) -> Weight:
    return Uniform(obj_name=obj_name, shape=shape, trainable=trainable)


def ones(obj_name: str, shape: List[int], trainable: Optional[bool]=None) -> Weight:
    return Ones(obj_name=obj_name, shape=shape, trainable=trainable)


def zeros(obj_name: str, shape: List[int], trainable: Optional[bool]=None) -> Weight:
    return Zeros(obj_name=obj_name, shape=shape, trainable=trainable)


def glorot_uniform(obj_name: str, shape: List[int], trainable: Optional[bool]=None) -> Weight:
    return Glorot(obj_name=obj_name, shape=shape, trainable=trainable)


import ospark.utility.weight_initializer
from ospark.nn.layers import *
import ospark


class EmbeddingLayer(Layer):

    def __init__(self,
                 obj_name: str,
                 embedding_dimension: int,
                 corpus_size: int,
                 is_training: Optional[bool]=None):
        super().__init__(obj_name=obj_name, is_training=is_training)
        self._embedding_dimension = embedding_dimension
        self._corpus_size         = corpus_size

    @property
    def embedding_dimension(self) -> int:
        return self._embedding_dimension

    @property
    def corpus_size(self) -> int:
        return self._corpus_size

    def in_creating(self) -> NoReturn:
        with tf.device("cpu:0"):
            self._embedding_layer = ospark.utility.weight_initializer.uniform(obj_name="embedding_layer",
                                                                              shape=[self.corpus_size,
                                                                 self.embedding_dimension])
            # self.assign(component=ospark.weight.uniform(obj_name="embedding_layer",
            #                                             weight_shape=[self.corpus_size,
            #                                                           self.embedding_dimension]))

    def pipeline(self, input_data: tf.Tensor) -> tf.Tensor:
        with tf.device("cpu:0"):
            sequence = tf.nn.embedding_lookup(self._embedding_layer, ids=input_data)
        return sequence
        # mask = tf.cast(tf.math.not_equal(input_data, 0), tf.float32)[:, :, tf.newaxis]
        # input_data = tf.one_hot(indices=input_data, depth=self.corpus_size) * mask
        # return tf.matmul(input_data, self.assigned.embedding_layer)


from ospark.nn.component.basic_module import ModelObject
from typing import NoReturn, Optional
import tensorflow as tf

class Backbone(ModelObject):

    def __init__(self,
                 obj_name: str,
                 is_keep_blocks: Optional[bool]=None,
                 is_training: Optional[bool]=None):
        super().__init__(obj_name=obj_name, is_training=is_training)
        if is_keep_blocks is not None:
            self._is_keep_blocks = is_keep_blocks
        else:
            self._is_keep_blocks = False
        self._block_register = []

    @property
    def is_keep_blocks(self) -> bool:
        return self._is_keep_blocks

    @property
    def block_register(self) -> list:
        return self._block_register

    def in_creating(self) -> NoReturn:
        raise NotImplementedError("Is required")

    def pipeline(self, input_data: tf.Tensor) -> tf.Tensor:
        raise NotImplementedError("Is required")


from ospark.nn.layers.feed_forward import *
from typing import NoReturn, Union
import tensorflow as tf


class DeepFeedForwardLayer(FeedForwardLayer):

    def __init__(self,
                 obj_name: str,
                 embedding_size: int,
                 scale_rate: int,
                 dropout_rate: float,
                 training_phase: Optional[bool]=None,
                 is_training: Optional[bool]=None,
                 activation: Optional[Activation]=None) -> NoReturn:
        super(DeepFeedForwardLayer, self).__init__(obj_name=obj_name,
                                                   embedding_size=embedding_size,
                                                   scale_rate=scale_rate,
                                                   activation=activation,
                                                   dropout_rate=dropout_rate,
                                                   is_training=is_training,
                                                   training_phase=training_phase)
        self._res_adaptive_value = 1

    @property
    def res_adaptive_value(self) -> Union[int, tf.Tensor]:
        return self._res_adaptive_value

    def profiling_phase(self, input_data: tf.Tensor, catch_variance: list):
        self._res_adaptive_value = 1.

        output   = self.__call__(input_data)
        _, variance = tf.nn.moments(output, axes=[0, 1])
        catch_variance.append(variance)

        self._res_adaptive_value = tf.sqrt(tf.reduce_sum(catch_variance, axis=0))[tf.newaxis, tf.newaxis, :]
        return output, catch_variance

    def residual_net(self, input_data: tf.Tensor) -> tf.Tensor:
        output = tf.multiply(input_data, self.res_adaptive_value)
        return output

    def back_to_standard(self) -> NoReturn:
        self._res_adaptive_value = 1.


from ospark.predictor import Predictor
from ospark.models.pixel_wise import PixelWiseDetection
from ospark.models.text_recognition import TextRecognition
from ospark.utility.roi_rotate import RoIRotate
from ospark.nn.component.weight import WeightOperator
from ospark.utility.non_max_suppression import NonMaxSuppression
from typing import Tuple, Optional, NoReturn
import tensorflow as tf
import numpy as np


class FOTSPredictor(Predictor):

    def __init__(self,
                 detection_model: PixelWiseDetection,
                 recognition_model: TextRecognition,
                 corpus: dict,
                 roi_image_height: Optional[int]=8,
                 score_threshold: Optional[float]=0.7,
                 iou_threshold: Optional[float]=0.5):
        self._detection_model   = detection_model
        self._recognition_model = recognition_model
        self._corpus            = corpus
        self._score_threshold   = score_threshold
        self._rotator           = RoIRotate(batch_size=1, target_height=roi_image_height)
        self._nms_processor     = NonMaxSuppression(iou_threshold=iou_threshold)

    @property
    def detection_model(self) -> PixelWiseDetection:
        return self._detection_model

    @property
    def recognition_model(self) -> TextRecognition:
        return self._recognition_model

    @property
    def score_threshold(self) -> float:
        return self._score_threshold

    @property
    def corpus(self) -> dict:
        return self._corpus

    @property
    def rotator(self) -> RoIRotate:
        return self._rotator

    @property
    def nms_processor(self) -> NonMaxSuppression:
        return self._nms_processor


    def predict(self, input_data: tf.Tensor):
        image_size = np.shape(input_data.numpy())[:2]
        bbox_prediction, feature_map = self.detection_model(input_data=input_data)

        bboxes_datasets = self.create_bboxes_datasets(bbox_prediction=bbox_prediction)
        nms_result      = self.nms_processor.start(bboxes_datasets=bboxes_datasets, image_size=image_size)
        if len(nms_result) == 0:
            raise ValueError(f"No prediction exceeds the threshold {self.score_threshold}")
        text_boxes, _   = zip(*map(self.rotator.textbox_from(image=tf.squeeze(feature_map)), nms_result))
        text_prediction = []
        for text_box in text_boxes:
            recognize_prediction = self.recognition_model(input_data=text_box[tf.newaxis, ...])
            recognize_prediction = tf.transpose(recognize_prediction, [1, 0, 2])
            prediction, _ = tf.nn.ctc_beam_search_decoder(inputs=recognize_prediction,
                                                       sequence_length=[tf.shape(recognize_prediction)[0]])
            word_indices = tf.sparse.to_dense(prediction[0]).numpy()[0]
            text_prediction.append("".join([self.corpus[i] for i in word_indices]))
        return nms_result, text_prediction

    def create_bboxes_datasets(self, bbox_prediction: Tuple[tf.Tensor, tf.Tensor, tf.Tensor]):
        score_map, bbox_map, angle_map = [prediction.numpy() for prediction in bbox_prediction]

        indices           = np.where(score_map > self.score_threshold)
        predict_bbox      = np.around(bbox_map[0, indices[1], indices[2], :]).astype(np.int32)
        vectors           = np.zeros(shape=[predict_bbox.shape[0], 4, 2])

        vectors[:, [0, 2], 1] = predict_bbox[:, [0, 2]] * [-1, 1]
        vectors[:, [1, 3], 0] = predict_bbox[:, [1, 3]] * [1, -1]
        predict_angle         = np.squeeze(angle_map[0, indices[1], indices[2], :])
        rotation_matrix       = np.concatenate(
                                [np.concatenate([np.cos(predict_angle)[..., np.newaxis, np.newaxis],
                                                 -np.sin(predict_angle)[..., np.newaxis, np.newaxis]], axis=-1),
                                 np.concatenate([np.sin(predict_angle)[..., np.newaxis, np.newaxis],
                                                 np.cos(predict_angle)[..., np.newaxis, np.newaxis]], axis=-1)],
                                axis=-2)
        rotate_vectors        = np.matmul(vectors, rotation_matrix)

        bbox_datasets = []
        for index in range(len(indices[0])):
            score = score_map[indices[0][index], indices[1][index], indices[2][index], indices[3][index]]
            reference_coordinate = np.array([indices[2][index], indices[1][index]])
            bbox_points = self.calculate_bbox_points(reference_coordinate=reference_coordinate, vectors=rotate_vectors[index, ...])
            bbox_datasets.append((score, np.ceil(bbox_points).astype(np.int32)))
        return bbox_datasets

    def calculate_bbox_points(self, reference_coordinate: np.ndarray, vectors: np.ndarray) -> np.ndarray:
        bbox_points = reference_coordinate + (vectors + vectors[[3, 0, 1, 2], :])
        return np.maximum(bbox_points, 0)

    def restore_weights(self, weights: dict) -> NoReturn:
        WeightOperator.restore(weights=weights)
        self.detection_model.create()
        self.recognition_model.create()



from ospark.data.generator.fots_data_generator import FOTSDataGenerator
from ospark.ocr.fots import FOTSTrainer
from ospark.nn.optimizer.learning_rate_schedule import TransformerWarmup
from ospark.data.folder import DataFolder
import tensorflow as tf
import json
import os

os_envi = "mac"

if os_envi == "mac":
    # dataset path
    folder_path          = "/Users/abnertsai/Documents/ICDAR/ch4/"
    training_data_folder = folder_path + "training_data"
    target_data_folder   = folder_path + "ch4_training_localization_transcription_gt"
    corpus_path          = folder_path + "corpus.json"

elif os_envi == "gpu":
    os.environ["CUDA_VISIBLE_DEVICES"] = "0"
    # dataset path
    folder_path          = "/home/ai/abner/ICDAR/ICDAR/ch4/"
    training_data_folder = folder_path + "training_data"
    target_data_folder   = folder_path + "ch4_training_localization_transcription_gt"
    corpus_path          = folder_path + "corpus.json"

elif os_envi == "maki_gpu":
    os.environ["CUDA_VISIBLE_DEVICES"] = "0"
    # dataset path
    folder_path          = os.path.dirname(os.path.abspath(__file__))
    training_data_folder = folder_path + "/dataset/ICDAR/training_data"
    target_data_folder   = folder_path + "/dataset/ICDAR/ch4_training_localization_transcription_gt"
    corpus_path          = folder_path + "/dataset/ICDAR/corpus.json"

# load corpus
with open(corpus_path, 'r') as fp:
    corpus = json.load(fp)

# get data folder
data_folder   = DataFolder(train_data_folder=training_data_folder, label_data_folder=target_data_folder)
# create data generator
data_generator = FOTSDataGenerator(data_folder=data_folder,
                                   batch_size=2,
                                   height_threshold=8,
                                   ignore_words={"###"},
                                   target_size=[1280, 720],
                                   image_shrink_scale=0.25)

save_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "fots_weight.json")
if os.path.isfile(save_path):
    with open(save_path, 'r') as fp:
        weight = json.load(fp)
else:
    weight = None

# create fots trainer
trainer = FOTSTrainer.create_attention_fots(data_generator=data_generator,
                                            corpus=corpus["char_to_index"],
                                            retrained_weights=weight,
                                            batch_size=16,
                                            epoch_number=50,
                                            optimizer=tf.keras.optimizers.Adam(learning_rate=TransformerWarmup(model_dimension=512)),
                                            reg_optimizer=tf.keras.optimizers.Adam(learning_rate=TransformerWarmup(model_dimension=512)),
                                            save_path=save_path
                                            )

# start training
trainer.start()



from .resnet import ResnetBackbone
from ospark.nn.cell.resnet_cell import resnet_cells


def resnet50(is_training: bool, catch_output: bool) -> ResnetBackbone:
    cells_number   = [3, 4, 6, 3]
    input_channels = [64, 256, 512, 1024]
    main_channels  = [64, 128, 256, 512]
    scale_rate     = 4
    cells          = resnet_cells(cells_number=cells_number,
                                  input_channels=input_channels,
                                  main_channels=main_channels,
                                  scale_rate=scale_rate,
                                  block_type="block_1",
                                  is_training=is_training)
    backbone = ResnetBackbone(obj_name="resnet_50",
                              cells=cells,
                              is_keep_blocks=catch_output,
                              is_training=is_training
                              )
    return backbone


def resnet101(is_training: bool, catch_output: bool) -> ResnetBackbone:
    cells_number   = [3, 4, 23, 3]
    input_channels = [64, 256, 512, 1024]
    main_channels  = [64, 128, 256, 512]
    scale_rate     = 4
    cells          = resnet_cells(cells_number=cells_number,
                                  input_channels=input_channels,
                                  main_channels=main_channels,
                                  scale_rate=scale_rate,
                                  block_type="block_1",
                                  is_training=is_training)
    backbone = ResnetBackbone(obj_name="resnet_101",
                              cells=cells,
                              is_keep_blocks=catch_output,
                              is_training=is_training
                              )
    return backbone


def resnet152(is_training: bool, catch_output: bool) -> ResnetBackbone:
    cells_number   = [3, 8, 36, 3]
    input_channels = [64, 256, 512, 1024]
    main_channels  = [64, 128, 256, 512]
    scale_rate     = 4
    cells          = resnet_cells(cells_number=cells_number,
                                  input_channels=input_channels,
                                  main_channels=main_channels,
                                  scale_rate=scale_rate,
                                  block_type="block_1",
                                  is_training=is_training)
    backbone = ResnetBackbone(obj_name="resnet_152",
                              cells=cells,
                              is_keep_blocks=catch_output,
                              is_training=is_training
                              )
    return backbone


def resnet18(is_training: bool, catch_output: bool) -> ResnetBackbone:
    cells_number   = [2, 2, 2, 2]
    main_channels  = [64, 128, 256, 512]
    input_channels = [64, 64, 128, 256]
    scale_rate     = 1
    cells          = resnet_cells(cells_number=cells_number,
                                  input_channels=input_channels,
                                  main_channels=main_channels,
                                  scale_rate=scale_rate,
                                  block_type="block_2",
                                  is_training=is_training)
    backbone = ResnetBackbone(obj_name="resnet_18",
                              cells=cells,
                              is_keep_blocks=catch_output,
                              is_training=is_training
                              )
    return backbone


def resnet34(is_training: bool, catch_output: bool) -> ResnetBackbone:
    cells_number   = [3, 4, 6, 3]
    main_channels  = [64, 128, 256, 512]
    input_channels = [64, 64, 128, 256]
    scale_rate     = 1
    cells          = resnet_cells(cells_number=cells_number,
                                  input_channels=input_channels,
                                  main_channels=main_channels,
                                  scale_rate=scale_rate,
                                  block_type="block_2",
                                  is_training=is_training)
    backbone = ResnetBackbone(obj_name="resnet_34",
                              cells=cells,
                              is_keep_blocks=catch_output,
                              is_training=is_training
                              )
    return backbone


from ospark import Model
from ospark.nn.block import Block
from ospark.nn.layers import Layer
from ospark.nn.layers.dense_layer import DenseLayer
from ospark.nn.block.vgg_block import fots_like_vgg
from ospark.nn.block.transformer_block import transformer_encoder_block
from typing import List, Optional, Callable
from functools import reduce
import tensorflow as tf


class TextRecognition(Model):

    def __init__(self,
                 obj_name: str,
                 classify_layer: Layer,
                 sequential_conv: List[Block],
                 sequential_model: List[Block],
                 trained_weights: Optional[dict]=None):
        super().__init__(obj_name=obj_name, trained_weights=trained_weights)
        self._classify_layer   = classify_layer
        self._sequential_conv  = sequential_conv
        self._sequential_model = sequential_model

    @property
    def classify_layer(self) -> Layer:
        return self._classify_layer

    @property
    def sequential_conv(self) -> List[Block]:
        return self._sequential_conv

    @property
    def sequential_model(self) -> List[Block]:
        return self._sequential_model

    # def in_creating(self) -> NoReturn:
    #     for layer in self.sequential_conv:
    #         self.assign(component=layer)
    #     for layer in self.sequential_model:
    #         self.assign(component=layer)

    def pipeline(self, input_data: tf.Tensor) -> tf.Tensor:
        mask = tf.cast(tf.math.not_equal(input_data[:, 0, :, 0], 0), dtype=tf.float32)
        sequential_conv_output = reduce(lambda output, layer: layer.pipeline(output), self.sequential_conv, input_data) * mask[:, tf.newaxis, :, tf.newaxis]
        sequence_model_output  = reduce(lambda output, layer: layer.pipeline(output), self.sequential_model, tf.squeeze(sequential_conv_output, axis=1))
        prediction             = self.classify_layer.pipeline(sequence_model_output)
        return prediction


def fots_recognition_model(class_number: int,
                           scale_rate: int,
                           head_number: int,
                           input_channel: int,
                           retrained_weights:dict,
                           sequential_output_channels: List[List[int]],
                           dropout_rate: Optional[float]=0.0,
                           sequential_model_block_number: Optional[int]=4,
                           trainable: Optional[bool]=True
                           ) -> TextRecognition:
    sequential_conv  = fots_like_vgg(input_channel=input_channel,
                                     output_channels=sequential_output_channels,
                                     trainable=trainable)
    sequential_model = []
    for i in range(sequential_model_block_number):
        name  = f"block_{i}"
        block = transformer_encoder_block(obj_name=name,
                                          embedding_size=sequential_output_channels[-1][-1],
                                          head_number=head_number,
                                          scale_rate=scale_rate,
                                          dropout_rate=dropout_rate,
                                          is_training=trainable)
        sequential_model.append(block)

    classify_layer   = DenseLayer(obj_name="classify_layer",
                                  input_dimension=sequential_output_channels[-1][-1],
                                  hidden_dimension=[class_number])
    return TextRecognition(obj_name="recognition_model",
                           trained_weights=retrained_weights,
                           sequential_conv=sequential_conv,
                           sequential_model=sequential_model,
                           classify_layer=classify_layer)


from tensorflow_datasets.core.deprecated.text.subword_text_encoder import SubwordTextEncoder


class LanguageDataEncoder:

    def __init__(self, train_data_encoder: SubwordTextEncoder, label_data_encoder: SubwordTextEncoder):
        self._train_data_encoder = train_data_encoder
        self._label_data_encoder = label_data_encoder

    @property
    def train_data_encoder(self) -> SubwordTextEncoder:
        return self._train_data_encoder

    @property
    def label_data_encoder(self) -> SubwordTextEncoder:
        return self._label_data_encoder

    def encode_train_data(self, input_data: list) -> list:
        return self.train_data_encoder.encode(input_data)

    def encode_label_data(self, input_data: list) -> list:
        return self.label_data_encoder.encode(input_data)


from ospark.nn.block import Block
from typing import NoReturn, Optional
from ospark.nn.layers.convolution_layer import ConvolutionLayer
import tensorflow as tf


class ResnetBlock(Block):

    def __init__(self,
                 obj_name: str,
                 input_channel: int,
                 main_channel: int,
                 scale_rate: Optional[int]=4,
                 kernel_size: Optional[int]=3,
                 strides: Optional[int]=1,
                 use_shortcut_conv: Optional[bool]=False,
                 is_training: Optional[bool]=True) -> NoReturn:
        super().__init__(obj_name=obj_name, is_training=is_training)
        self._input_channel  = input_channel
        self._main_channel   = main_channel
        self._scale_rate     = scale_rate
        self._kernel_size    = kernel_size
        self._strides        = strides

        self._use_shortcut_conv  = use_shortcut_conv

    @property
    def input_channel(self) -> int:
        return self._input_channel

    @property
    def main_channel(self) -> int:
        return self._main_channel

    @property
    def scale_rate(self) -> int:
        return self._scale_rate

    @property
    def kernel_size(self) -> int:
        return self._kernel_size

    @property
    def strides(self) -> int:
        return self._strides

    @property
    def use_shortcut_conv(self) -> bool:
        return self._use_shortcut_conv

    def in_creating(self) -> NoReturn:
        if self.use_shortcut_conv:
            self.shortcut_conv = ConvolutionLayer.bn_relu_conv(obj_name="shortcut_conv",
                                                               filter_size=[1,
                                                                            1,
                                                                            self.input_channel,
                                                                            self.scale_rate * self.main_channel],
                                                               strides=[1, self.strides, self.strides, 1],
                                                               padding="SAME",
                                                               is_training=self.is_training)

    def pipeline(self, input_data: tf.Tensor) -> tf.Tensor:
        if self.use_shortcut_conv:
            shortcut = self.shortcut_conv.pipeline(input_data)
        else:
            shortcut = input_data
        return shortcut


class BottleneckBuildingBlock(ResnetBlock):

    def in_creating(self) -> NoReturn:
        super().in_creating()
        if self.use_shortcut_conv:
            self.layer_1X1_0 = ConvolutionLayer.conv_bn_relu(obj_name="layer_1X1_0",
                                                             filter_size=[1,
                                                                          1,
                                                                          self.input_channel,
                                                                          self.main_channel],
                                                             strides=[1, self.strides, self.strides, 1],
                                                             padding="SAME",
                                                             is_training=self.is_training)
        else:
            self.layer_1X1_0 = ConvolutionLayer.conv_bn_relu(obj_name="layer_1X1_0",
                                                             filter_size=[1,
                                                                          1,
                                                                          self.scale_rate * self.main_channel,
                                                                          self.main_channel],
                                                             strides=[1, self.strides, self.strides, 1],
                                                             padding="SAME",
                                                             is_training=self.is_training)

        self.layer_3X3 = ConvolutionLayer.conv_bn_relu(obj_name="layer_3X3",
                                                       filter_size=[self.kernel_size,
                                                                    self.kernel_size,
                                                                    self.main_channel,
                                                                    self.main_channel],
                                                       strides=[1, 1, 1, 1],
                                                       padding="SAME",
                                                       is_training=self.is_training)
        self.layer_1X1_1 = ConvolutionLayer.conv_bn_relu(obj_name="layer_1X1_1",
                                                         filter_size=[1,
                                                                      1,
                                                                      self.main_channel,
                                                                      self.scale_rate * self.main_channel],
                                                         strides=[1, 1, 1, 1],
                                                         padding="SAME",
                                                         is_training=self.is_training)

    def pipeline(self, input_data: tf.Tensor) -> tf.Tensor:
        output = input_data
        shortcut = super().pipeline(input_data)
        layers = [self.layer_1X1_0, self.layer_3X3, self.layer_1X1_1]
        for layer in layers:
            output = layer.pipeline(output)
        return output + shortcut

class BuildingBlock(ResnetBlock):

    def in_creating(self) -> NoReturn:
        super().in_creating()
        if self.use_shortcut_conv:
            self.layer_3X3_0 = ConvolutionLayer.conv_bn_relu(obj_name="layer_3X3_0",
                                                             filter_size=[3,
                                                                          3,
                                                                          self.input_channel,
                                                                          self.main_channel],
                                                             strides=[1, self.strides, self.strides, 1],
                                                             padding="SAME",
                                                             is_training=self.is_training)
        else:
            self.layer_3X3_0 = ConvolutionLayer.conv_bn_relu(obj_name="layer_3X3_0",
                                                             filter_size=[3,
                                                                          3,
                                                                          self.scale_rate * self.main_channel,
                                                                          self.main_channel],
                                                             strides=[1, self.strides, self.strides, 1],
                                                             padding="SAME",
                                                             is_training=self.is_training)
        self.layer_3X3_1 = ConvolutionLayer.conv_bn_relu(obj_name="layer_3X3_1",
                                                         filter_size=[3,
                                                                      3,
                                                                      self.main_channel,
                                                                      self.scale_rate * self.main_channel],
                                                         strides=[1, 1, 1, 1],
                                                         padding="SAME",
                                                         is_training=self.is_training)

    def pipeline(self, input_data: tf.Tensor) -> tf.Tensor:
        output = input_data
        shortcut = super().pipeline(input_data)
        layers = [self.layer_3X3_0, self.layer_3X3_1]
        for layer in layers:
            output = layer.pipeline(output)
        return output + shortcut




from typing import Optional

import tensorflow as tf

from ospark.nn.loss_function.loss_function import LossFunction


class Degree(LossFunction):

    def __init__(self, coefficient: Optional[float]=1.0):
        self._coefficient = coefficient

    @property
    def coefficient(self) -> float:
        return self._coefficient

    def calculate(self,
                  prediction: tf.Tensor,
                  target_data: tf.Tensor) -> tf.Tensor:
        cls_target = tf.cast(tf.where(target_data != 0), tf.float32)
        loss = (1.0 - tf.math.cos(prediction * cls_target - target_data))
        degree_loss = self.coefficient * tf.divide(tf.reduce_sum(loss), (tf.reduce_sum(cls_target) + 1))
        return degree_loss

    def __call__(self,
                 prediction: tf.Tensor,
                 target_data: tf.Tensor) -> tf.Tensor:
        loss_value = self.calculate(prediction=prediction, target_data=target_data)
        return loss_value